#%%
import timeit




def filter_data(data, lower, upper):
    result = []

    x, y = data.shape[0], len(lower)
    
    for i in range(x):
        keep = True
        for j in range(y):
            val = data[i, j]
            if val < lower[j]:
                keep = False
                break
            if val > upper[j]:
                keep = False
                break
        if keep:
            result.append(data[i])
    return np.array(result)

#%%


import timeit
import math
import numpy as np


def expansion(v1, p1, k2):
    if v1[0] > 0 and v1[1] == 0:
        angle = math.atan2(k2[1] - p1[1], k2[0] - p1[0])

    elif v1[0] < 0 and v1[1] == 0:
        angle = math.atan2(-(k2[1] - p1[1]), -(k2[0] - p1[0]))

    elif v1[0] == 0 and v1[1] > 0:
        angle = math.atan2(-(k2[0] - p1[0]), k2[1] - p1[1])

    elif v1[0] == 0 and v1[1] < 0:
        angle = math.atan2(k2[0] - p1[0], -(k2[1] - p1[1]))
    clockwise_angle = (-angle) % (2 * math.pi)
    return math.degrees(clockwise_angle)



def clockwise_angle(p1, p2, k2):
    v1 = p2 - p1
    v2 = k2 - p1

    dot = np.dot(v1, v2)
    det = v1[0]*v2[1] - v1[1]*v2[0]  
    angle_rad = math.atan2(det, dot) 
    clockwise_angle = (-angle_rad) % (2 * np.pi)
    return math.degrees(clockwise_angle)

p1 = (5, 20)
p2 = (5, 0)

k2 = (18, 30)
v1 = (0, -20)

res = expansion(v1, p1, k2)
print(res)
res = clockwise_angle(np.array(p1), np.array(p2), np.array(k2))
print(res)


#%%

import timeit

setup1 = """
from dataclasses import dataclass
import numpy as np

@dataclass(slots=True, frozen=False)
class OptionGreeks:
    delta: float = np.nan
    gamma: float = np.nan
    vega: float = np.nan
    theta: float = np.nan
    rho: float = np.nan

og = OptionGreeks(0.5, 0.1, 0.2, -0.3, 0.4)
"""

run1 = """
og.vega
"""

setup2 = """
from dataclasses import dataclass
import numpy as np

@dataclass(slots=True, frozen=False)
class OptionGreeks:
    delta: float = np.nan
    gamma: float = np.nan
    _vega: float = np.nan
    theta: float = np.nan
    rho: float = np.nan
    
    @property
    def vega(self):return self._vega

og = OptionGreeks(0.5, 0.1, 0.2, -0.3, 0.4)
"""

run2 = """
og.vega
"""

setup3 = """
from dataclasses import dataclass
import numpy as np

@dataclass(slots=True, frozen=True)
class OptionGreeks:
    delta: float = np.nan
    gamma: float = np.nan
    vega: float = np.nan
    theta: float = np.nan
    rho: float = np.nan

og = OptionGreeks(0.5, 0.1, 0.2, -0.3, 0.4)
"""

run3 = """
og.__getattribute__("vega")
"""



setup4 = """
greeks_dict = {
    'delta': 0.5,
    'gamma': 0.1,
    'vega': 0.2,
    'theta': -0.3,
    'rho': 0.4
}
"""

run4 = """
greeks_dict["vega"]
"""
from dataclasses import dataclass

@dataclass(slots=True)
class Greeks:
    _ptype_idx_map: dict
    _delta: float
    _gamma: float
    _vega: float
    _theta: float
    _rho: float

def make_property(attr_name):
    def prop(self):
        return getattr(self, f"_{attr_name}")
    return property(prop)

for var in ("delta", "gamma", "vega", "theta", "rho"):
    setattr(Greeks, var, make_property(var))

setup5 = """
from __main__ import Greeks
g = Greeks({}, 0.5, 0.1, 0.2, -0.3, 0.4)
"""

run5 = """
g.vega
"""

N = 50000000
t1 = timeit.timeit(run1, setup=setup1, number=N)
t2 = timeit.timeit(run2, setup=setup2, number=N)
#t3 = timeit.timeit(run3, setup=setup3, number=N)
#t4 = timeit.timeit(run4, setup=setup4, number=N)
#t5 = timeit.timeit(run5, setup=setup5, number=N)

print(f"Direct access (slots): {t1:.6f} seconds")
print(f"getattr access (slots): {t2:.6f} seconds")
print(f"getattr access (slots+frozen): {t2:.6f} seconds")
print(f"Dict access: {t4:.6f} seconds")
print(f"Direct access (slots+frozen): {t5:.6f} seconds")

#%%
"""
from py_lets_be_rational.normaldistribution import norm_cdf
import time
discounted_option_price= 10
F=100.
K=105.
r=0.05
t=1
flag="c"

N = 1000000

start1 = time.time()

for _ in range(N):
    for _ in range(2):
        norm_cdf(0.5)

print(time.time() - start1)

import py_vollib_vectorized 
import numpy as np


price= 10
F=100.
K=105.
r=0.05
t=1
flag="c"

from scipy.special import ndtr

start2 = time.time()

for _ in range(N):
    for _ in range(2):
        ndtr(0.5)

print(time.time() - start2)



from scipy.stats import norm

from scipy.special import ndtr

start2 = time.time()
arr = np.array([0.5, 0.6])
for _ in range(N):
    norm.cdf(arr, 0, 1)

print(time.time() - start2)
"""
# %%

#$import numpy as np
#import py_vollib_vectorized




from pprint import pprint
S = 100.
K=102.
t=0.002
r=0.05
q=0.02
flag="c"
iv=0.2

#greeks = py_vollib_vectorized.get_all_greeks(flag, S, K, t, r, iv, model='black_scholes', return_as='dict')
#pprint(greeks)


#%%

import timeit

setup1 = """

d = {i : i for i in range(10)}



"""

run1 = """


for i, j in d.items():
    pass


"""

setup2 = """
d = {i : i for i in range(10)}


"""

run2 = """

for i in d:
    v=d[i]


"""


N = 10000000
t1 = timeit.timeit(run1, setup=setup1, number=N)
t2 = timeit.timeit(run2, setup=setup2, number=N)



print(f"t1: {t1:.6f} seconds")
print(f"t2: {t2:.6f} seconds")

#%%


class StaticClass:
    @staticmethod
    def delta(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        return flag * math.exp(-r*t)*norm_cdf(flag*d1)

    @staticmethod
    def gamma(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        return math.exp(-r * t) * norm_cdf(d1) / (F * sigma * math.sqrt(t))

    @staticmethod
    def vega(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        return F * math.exp(-r*t) * norm_cdf(d1, ) * math.sqrt(t)

    @staticmethod
    def theta(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        d2 = d1 - math.sqrt(t)
        t1 = - F * math.exp(-r*t) * norm_pdf(d1) * sigma / (2 * math.sqrt(t))
        t2 = - flag * r * K *math.exp(-r*t) * norm_cdf(flag*d2) + flag* r * F * math.exp(-r*t) * norm_cdf(flag*d1)
        return t1 + t2

    @staticmethod
    def rho(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        d2 = d1 - math.sqrt(t)
        return -t * flag * (F * norm_cdf(flag *sigma) - K * norm_cdf(flag*d2))

    @staticmethod
    def price(sigma, F, K, t, r, flag, **kwargs):
        d1 = F * norm_cdf(flag * (math.log(F / K) / sigma + sigma/2))
        d2 = K * norm_cdf(flag * (math.log(F / K) / sigma - sigma/2))
        return math.exp(-r*t) * (d1 - d2)



# Add this timing test
def time_static_class():
    results = {}
    for greek in greeks:
        fn = getattr(StaticClass, greek)
        results[greek] = fn(**params)
    return results

N = 20000


# Test setup
params = {
    'sigma': 0.2,
    'F': 100,
    'K': 100,
    't': 1,
    'r': 0.05,
    'flag': 1
}
greeks = {'delta', 'gamma', 'vega', 'theta', 'rho', 'price'}


greeks =  {'delta', 'gamma', 'vega', 'theta', 'rho', 'price'}

original = BlackScholesCacheOriginal(**params)
new = BlackScholesCacheNew(**params)

original_time = timeit.timeit(lambda: original.build_cache(greeks), number=N)
new_time = timeit.timeit(lambda: new.build_cache(greeks), number=N)
#static_time = timeit.timeit(time_static_class, number=N)


print(f"Original: {original_time:.5f} seconds")
print(f"New: {new_time:.5f} seconds") 
print(f"Static: {static_time:.5f} seconds")

# %%
# %%

import math
from dataclasses import dataclass
from typing import Dict, Set, Callable
from scipy.stats import norm




@dataclass(frozen=True)
class PricingModelCache:
    sigma: float
    F: float
    K: float
    t: float
    r: float
    flag: int
    model_type: str = "black76"  

    def build_cache(self, greeks: Set[str]) -> Dict[str, float]:
        return IntermediateCalculations.calculate(
            greeks=greeks,
            F=self.F,
            K=self.K,
            sigma=self.sigma,
            t=self.t,
            r=self.r,
            flag=self.flag,
            model_type=self.model_type
        )

@dataclass(frozen=True)
class IntermediateCalculations:
    _calculation_methods = {
        "black_scholes": {
            "d1": lambda F, K, sigma, t, r, **_: (math.log(F / K) + (r + 0.5 * sigma**2) * t) / (sigma * math.sqrt(t)),
            "d2": lambda d1, sigma, t, **_: d1 - sigma * math.sqrt(t),
        },
        "black76": {
            "d1": lambda F, K, sigma, t, **_: (math.log(F / K) + 0.5 * sigma**2 * t) / (sigma * math.sqrt(t)),
            "d2": lambda d1, sigma, t, **_: d1 - sigma * math.sqrt(t),
        }
    }

    _common_methods = {
        "phi_d1": lambda d1, **_: norm.pdf(d1),
        "phi_d2": lambda d2, **_: norm.pdf(d2),
        "N_d1": lambda d1, **_: norm.cdf(d1),
        "N_d2": lambda d2, **_: norm.cdf(d2),
        "N_flag_d1": lambda d1, flag, **_: norm.cdf(flag * d1),
        "N_flag_d2": lambda d2, flag, **_: norm.cdf(flag * d2),
    }

    @classmethod
    def calculate(cls, greeks, F, K, sigma, t, r, flag, model_type):
        greek_requirements = {
            "delta": {"d1", "N_flag_d1"},
            "gamma": {"d1", "phi_d1"},
            "vega": {"d1", "phi_d1"},
            "theta": {"d1", "d2", "phi_d1", "N_flag_d1", "N_flag_d2"},
            "rho": {"d2", "N_flag_d2"},
            "price": {"d1", "d2", "N_flag_d1", "N_flag_d2"},
        }

        required = set()
        for g in greeks:
            required |= greek_requirements.get(g, set())

        cache = {}
        params = {'F': F, 'K': K, 'sigma': sigma, 't': t, 'r': r, 'flag': flag}
        
        if required and "d1" in required:
            cache["d1"] = cls._calculation_methods[model_type]["d1"](**params)
            params.update(cache)

        calculation_order = ["d2", "phi_d1", "phi_d2", "N_d1", "N_d2", "N_flag_d1", "N_flag_d2"]
        for key in calculation_order:
            if key in required and key not in cache:
                if key in cls._calculation_methods[model_type]:
                    cache[key] = cls._calculation_methods[model_type][key](**params)
                else:
                    cache[key] = cls._common_methods[key](**params)
                params[key] = cache[key]

        return cache


import math
import timeit
from dataclasses import dataclass
from scipy.stats import norm
import numpy as np
norm_cdf = norm.cdf
norm_pdf = norm.pdf
# Original implementation
@dataclass(frozen=True)
class BlackScholesCacheOriginal:
    sigma: float
    F: float
    K: float
    t: float
    r: float
    flag: int

    def build_cache(self, greeks: set[str]) -> dict:
        cache = {}
        sigma_sqrt_t = self.sigma * math.sqrt(self.t)
        d1 = (math.log(self.F / self.K) + 0.5 * self.sigma**2 * self.t) / sigma_sqrt_t
        d2 = d1 - sigma_sqrt_t

        requirements = {
            "delta": {"d1", "N_flag_d1"},
            "gamma": {"d1", "phi_d1"},
            "vega": {"d1", "phi_d1"},
            "theta": {"d1", "d2", "phi_d1", "N_flag_d1", "N_flag_d2"},
            "rho": {"d2", "N_flag_d2"},
            "price": {"d1", "d2", "N_flag_d1", "N_flag_d2"},
        }

        required = set()
        for g in greeks:
            required |= requirements.get(g, set())

        if "d1" in required:
            cache["d1"] = d1
        if "d2" in required:
            cache["d2"] = d2
        if "phi_d1" in required:
            cache["phi_d1"] = norm.pdf(d1)
        if "phi_d2" in required:
            cache["phi_d2"] = norm.pdf(d2)
        if "N_d1" in required:
            cache["N_d1"] = norm.cdf(d1)
        if "N_d2" in required:
            cache["N_d2"] = norm.cdf(d2)
        if "N_flag_d1" in required:
            cache["N_flag_d1"] = norm.cdf(self.flag * d1)
        if "N_flag_d2" in required:
            cache["N_flag_d2"] = norm.cdf(self.flag * d2)

        return cache



class StaticClass:
    @staticmethod
    def delta(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        return flag * math.exp(-r*t)*norm_cdf(flag*d1)

    @staticmethod
    def gamma(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        return math.exp(-r * t) * norm_pdf(d1) / (F * sigma * math.sqrt(t))

    @staticmethod
    def vega(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        return F * math.exp(-r*t) * norm_pdf(d1) * math.sqrt(t)

    @staticmethod
    def theta(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        d2 = d1 - sigma*math.sqrt(t)
        t1 = - F * math.exp(-r*t) * norm_pdf(d1) * sigma / (2 * math.sqrt(t))
        t2 = - flag * r * K *math.exp(-r*t) * norm_cdf(flag*d2) + flag* r * F * math.exp(-r*t) * norm_cdf(flag*d1)
        return t1 + t2

    @staticmethod
    def rho(sigma, F, K, t, r, flag, **kwargs):
        d1 =  1/(sigma*math.sqrt(t)) * (np.log(F/K) + (0.5*sigma**2)*t)
        d2 = d1 - sigma*math.sqrt(t)
        return flag * K * t * math.exp(-r*t) * norm_cdf(flag*d2)

    @staticmethod
    def price(sigma, F, K, t, r, flag, **kwargs):
        d1 = (math.log(F/K) + (r + 0.5*sigma**2)*t) / (sigma*math.sqrt(t))
        d2 = d1 - sigma*math.sqrt(t)
        return flag * (F * norm_cdf(flag*d1) - K * math.exp(-r*t) * norm_cdf(flag*d2))

N=10000

greeks = {'delta', 'gamma', 'vega', 'theta', 'rho', 'price'}


# Test setup
params = {
    'sigma': 0.2,
    'F': 100,
    'K': 100,
    't': 1,
    'r': 0.05,
    'flag': 1
}
pmc = PricingModelCache(**params)


original = BlackScholesCacheOriginal(**params)


original_time = timeit.timeit(lambda: original.build_cache(greeks), number=N)
new_time = timeit.timeit(lambda: pmc.build_cache(greeks), number=N)
#static_time = timeit.timeit(time_static_class, number=N)


print(f"Original: {original_time:.5f} seconds")
print(f"New: {new_time:.5f} seconds") 



from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING
    
import pyqtgraph as pg
from pyqtgraph import opengl 
import numpy as np
from typing import List, Optional
from PySide6 import QtCore, QtGui
from pyqtgraph import functions as fn
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from pyqtgraph import opengl
from PySide6.QtWidgets import QApplication, QMainWindow
from PySide6 import QtWidgets
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING
import sys
from PySide6 import QtGui, QtCore
import pyqtgraph.opengl as gl
import numpy as np
from pyqtgraph.opengl import GLTextItem
from OpenGL import GL as opengl
from OpenGL import GLU
from PySide6 import QtGui
from pprint import pprint
import copy
import math
from pyvol_terminal.interfaces.volatility_surface import view_box


class WorldSizedText(GLTextItem):
    world_width_character = 0.014286

    def __init__(self, *args, **kwargs):
        self.pos_arr = np.array(kwargs["pos"])
        self._reference_size=None
        self._right_anchor_pos=np.array(kwargs["pos"])
        super().__init__(*args, **kwargs)

        
    def setParent(self, parent: CustomView):
        super().setParent(parent)
        self.updatePixelSize(parent)
        
    @QtCore.Slot()
    def updatePixelSize(self, view_widget: CustomView):
        print(f"\n{self.text}")
        self.blockSignals(True)
        self.maintain_text_world_width()
        self.shiftToWidget(view_widget)
        self.blockSignals(False)
    
    def shiftToWidget(self, widget: CustomView):
        text_width_world = self.worldWidth()
        camera_pos = widget.cameraPositionNumpy()
        current_angle = math.degrees(math.atan2(*(-1*camera_pos[:2])))
        
        x, y = self.offset_xy(text_width_world, current_angle, 0)
        pos = self._right_anchor_pos.copy()
        pos[0] = pos[0] + x
        pos[1] = pos[1] + y
        self.setData(pos = pos)

    def offset_xy(self, txt_width, theta, elevation):
        x = -txt_width * math.cos(math.radians(theta))
        y = txt_width * math.sin(math.radians(theta))
        return x, y

    def maintain_text_world_width(self):
        current_world_width = self.worldWidth()
        if current_world_width <= 0:
            return
        
        scale_factor = len(self.text) * self.world_width_character / current_world_width
        
        current_font = self.font
        current_point_size = current_font.pointSizeF()
        new_size = current_point_size * scale_factor
        current_font.setPointSizeF(new_size)
        self.setData(font=current_font)

    def setData(self, *args, **kwargs):
        super().setData(*args, **kwargs)
        if "pos" in kwargs:
            self.pos_arr[:] = kwargs["pos"]
    
    def pxHeight(self):
        metrics = QtGui.QFontMetricsF(self.font)
        return metrics.boundingRect(self.text).height()
    
    def pxWidth(self):
        metrics = QtGui.QFontMetricsF(self.font)
        return metrics.boundingRect(self.text).width()
    
    def pointWidth(self):
        return self.pxWidth() * 72 / self.view().screen().logicalDotsPerInch()

    def worldWidth(self):
        return self.pxWidth() * self.view().pixelSize(self.pos_arr)

    def worldWidthPoint(self):
        return self.pxWidth() * 72 / self.view().screen().logicalDotsPerInch()#* self.view().pointSize(self.pos_arr)
    
    
class CustomView(gl.GLViewWidget):
    viewChangedSignal = QtCore.Signal(object)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.opts["azimuth"] = 0 
        self.opts["center"] = QtGui.QVector3D(0, 0, 0)
        self.opts["distance"] = 3
        self.opts["elevation"] = 0
        self.update()
        self._update_timer = QtCore.QTimer()
        self._update_timer.setSingleShot(True)
        self._update_timer.setInterval(50) 
        self._update_timer.timeout.connect(self._emit_view_changed)
        self._pending_view_change = False
        

    def cameraPositionNumpy(self):
        p = self.cameraPosition()
        return np.array((p.x(), p.y(), p.z()))
        
    def wheelEvent(self, ev):
        distance_prev = self.opts["distance"]
        super().wheelEvent(ev)  
        if self.opts["distance"] != distance_prev:
            self.update()
            self.viewChangedSignal.emit(self)
            
    def pointSize(self, pos):
        PIXELS_PER_POINT = self.screen().logicalDotsPerInch() / 72.0
        return PIXELS_PER_POINT * self.pixelSize(pos)

    def _emit_view_changed(self):
        self.viewChangedSignal.emit(self)
        self._pending_view_change = False

    def orbit(self, azim, elev):
        super().orbit(azim, elev)
        if not self._pending_view_change:
            self._pending_view_change = True
            self._update_timer.start()
            
            
class MainWindow(QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('Volatility Surface')
        self.widget_central = QtWidgets.QWidget()
        self.setCentralWidget(self.widget_central)
        self.layout_main = QtWidgets.QVBoxLayout()
        self.widget_central.setLayout(self.layout_main)
        self.showMaximized()

        self.view_widget = CustomView()
        self.vw_layout = QtWidgets.QVBoxLayout()
        self.layout_main.addLayout(self.vw_layout)
        
        self.layout_main.addWidget(self.view_widget)
        
        self.showMaximized()
        QtCore.QTimer.singleShot(100, self.init_axes_after_show)
        
    def init_axes_after_show(self):

        grid = gl.GLGridItem()
        grid.setSize(1, 1, 1)
        
        self.view_widget.addItem(grid)
        grid.translate(0.5, 0.5, 0)
        font=QtGui.QFont('Neue Haas Grotesk')#, pointSize=12)
        font.setPointSizeF(12)
        print(font.pointSizeF())
        positions = [[100000, "abc"], [120000, 132000]]
        for i in np.linspace(0, 1, 4):
            for j in np.linspace(0, 1, 4):
                if i % 1 == 0 or j % 1 == 0:
                    if i != 1:
                        continue
                    pos = (i, j, 0)
                    text = str(positions[int(i)][int(j)])#f"{(float(i),float(j))}"
                    text = f"{(round(float(i),2), round(float(j),2))}"
                    font=QtGui.QFont('Neue Haas Grotesk')#, pointSize=12)
                    font.setPointSizeF(12)

                    text_item = WorldSizedText(pos=pos, text=text, font=font)
                    self.view_widget.addItem(text_item)
                    self.view_widget.viewChangedSignal.connect(text_item.updatePixelSize)
                    text_item.setParent(self.view_widget)
        

if __name__ == "__main__":
    app = QApplication(sys.argv)
    mainWin = MainWindow()
    sys.exit(app.exec())
    



from __future__ import annotations
from typing import List, Dict, Union, Tuple, Callable, TYPE_CHECKING, Any, Optional, Any
if TYPE_CHECKING:
    from instruments.utils import InstrumentManager

from PySide6 import QtCore, QtWidgets, QtGui
from .. import windows
from pyvol_terminal import workers
from pyvol_terminal.data_classes import builders as builders_data_classes
from . import frontend, stylesheets
from ..workers import ABCCalibrationWorker
from ..interfaces.abstract_classes import ABCInterface
from ..requirements_tracker import RequirementsTracker


from PySide6 import QtCore, QtGui, QtWidgets
from pprint import pprint
from pyvol_terminal.interfaces.abstract_classes import ABCInterface

interface_container: Dict[str, ABCInterface] = {}

class Application(QtWidgets.QApplication):  
    create_window_signal = QtCore.Signal(dict, str, tuple)
    window_titles_signal = QtCore.Signal(list)
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance
    
    @classmethod
    def get_instance(cls):
        return cls._instance
    
    def __init__(self,
                 argv: List[str],
                 interface_config_container: Dict[str, Dict[str, Any]]=None,
                 ws_worker_config=None,
                 data_processing_config=None,
                 option_configurations: Dict[str, Any]=None,
                 requirements_tracker=None,
                 ):
        super().__init__(argv)
        
        self.option_configurations=option_configurations
        self.interface_configs=interface_config_container
        self.aboutToQuit.connect(self.exiting)
        self.n_windows=0
        self._calibration_workers: Dict[str, ABCCalibrationWorker]={}
        self._initialized = True
        self.requirements_tracker=requirements_tracker if not requirements_tracker is None else RequirementsTracker()
        
        self.setFont(stylesheets.get_global("font"))
        self.setStyleSheet(stylesheets.get_global("widgets"))

        main_instruments = list(option_configurations.keys())
        interfaces = list(interface_config_container.keys())
        
        self.terminal = frontend.Frontend(self.open_interface,
                                          main_instruments,
                                          interfaces
                                          )
        self.terminal.show()
        all_instruments = {}
        for option_config in self.option_configurations.values():
            all_instruments = all_instruments | option_config["instrument_manager"].all_instrument_objects
     
        self._websocket_worker = workers.WebsocketWorker(data_processing_config=data_processing_config,
                                                         ws_worker_config=ws_worker_config,
                                                         all_instruments=all_instruments,
                                                         )
        
        self.terminal.combo_box.setCurrentIndex(1)
        self.terminal.buttons[0].click()
 
    def start(self):
        self._websocket_worker.start()

    def open_interface(self, instrument: str, interface_type: str):        
        window_title = f"{instrument}: {interface_type}"
                
        option_chain_container = self.option_configurations[instrument]["options_chain_container"]
        instrument_manager: InstrumentManager = self.option_configurations[instrument]["instrument_manager"]
        
        active_instruments = set(instrument_manager.all_instrument_objects.keys())
        
        vol_vect_container = builders_data_classes.create_vol_vect_container(option_chain_container)
    
        surface_calibration_config = self.option_configurations[instrument]["surface_calibration_config"]
        
        if not instrument in self._calibration_workers:
            surface_calibration_worker=workers.SurfaceCalibration(option_chain_container=option_chain_container,
                                                                  vol_vect_container=vol_vect_container,
                                                                  calibration_engines=surface_calibration_config["engine_container"],
                                                                  active_instruments=active_instruments,
                                                                  update_timer=surface_calibration_config["update_timer"],
                                                                  )
            self._calibration_workers[instrument] = surface_calibration_worker
        else:
            surface_calibration_worker = self._calibration_workers[instrument]
        
        configs_to_add = {"instrument_manager" : instrument_manager,
                       #   "surface_calibration_config" : surface_calibration_config,
                          "vol_vect_container" : vol_vect_container,
                          }
        
        interface_configs = self.interface_configs[interface_type]
        interface_configs.update(configs_to_add)
        interface_class = interface_configs.get("widget_class")
        interface: ABCInterface = interface_class(**self.interface_configs[interface_type])
                
        interface_update_slot = interface.get_calibration_slot()
        surface_calibration_worker.calibratedSignal.connect(interface_update_slot)
        interface.aboutToCloseSig.connect(self.interface_closing)
        interface.setWindowTitle(window_title)
      #  self.connectInteraction(interface.main_view, [self._websocket_worker.setIntenseInteraction,
     #                                                 self._websocket_worker.setIntenseInteraction
     #                                                 ])
        
        interface_container[window_title] = interface
        self._websocket_worker.add_active_instruments(active_instruments)
                
    @QtCore.Slot(ABCInterface)
    def interface_closing(self, interface_closing: windows.InterfaceWindow):
        del interface_container[interface_closing.windowTitle()]

        current_active_instruments = set()
        delete_calibration_worker=True
        for remaining_interfaces in interface_container.values():
            if isinstance(remaining_interfaces, ABCInterface):
                calibration_worker = remaining_interfaces.calibration_worker
                current_active_instruments.update(calibration_worker.active_instruments)
                if calibration_worker == interface_closing.calibration_worker:
                    delete_calibration_worker = False
                    
        if delete_calibration_worker:
            interface_closing.calibration_worker.stop_worker()
            del self._calibration_workers[interface_closing.primary_instrument]

        self._websocket_worker.reset_active_instruments(current_active_instruments)
    
    def connectInteraction(self, main_view, receivers: List[Callable]):
        return
        for receiver in receivers:
            main_view.add_interact_end_callback(receiver)
                
        
    def open_sub_window(self,
                        widget: QtWidgets.QWidget,
                        geometry: QtCore.QRect=None,
                        adjustToContents: bool=False,
                        flags=None,
                        parent: QtWidgets.QWidget=None,
                        ) -> windows.SubWindow:
        
        for i in range(1, 9999):
            window_title = f"Sub Window {i}"
            if not window_title in interface_container:
                break 
        win = windows.SubWindow(widget, geometry, adjustToContents, flags, parent=parent)
        interface_container[window_title] = win

    @QtCore.Slot()
    def exiting(self):
        try:
            loop = self._websocket_worker.websocket.get_loop()
            import asyncio
            asyncio.run_coroutine_threadsafe(self._websocket_worker.websocket.stop(), loop)
            

        except:
            pass
        self._websocket_worker.stop_worker()

    def update_instrument_quantities(self, metric):
        self.requirements_tracker.add_interface_requirement(metric)
        print(self.requirements_tracker.current_requirements)
        
    def verify_object_method_returns_callable(self,
                                              obj: Any,
                                              method_name: str,
                                              *method_args,
                                              **method_kwargs
                                              ) -> tuple[bool, Optional[Callable]]:
        """
        Verify an instantiated object has a method that returns a callable.
        
        Args:
            obj: Instantiated object to check
            method_name: Name of the method to verify
            *method_args: Positional arguments to pass to the method
            **method_kwargs: Keyword arguments to pass to the method
            
        Returns:
            tuple: (success: bool, returned_callable: Optional[Callable])
                - success: True if verification passed
                - returned_callable: The callable if verification succeeded, else None
        """
        # Check if object has the method
        if not hasattr(obj, method_name):
            return (False, None)
        
        method = getattr(obj, method_name)
        
        # Check if it's a callable method
        if not callable(method):
            return (False, None)
        
        try:
            # Call the method and get its return value
            returned_value = method(*method_args, **method_kwargs)
            
            # Verify the returned value is callable
            if callable(returned_value):
                return (True, returned_value)
            return (False, None)
            
        except Exception as e:
            # Handle any exceptions during method call
            print(f"Verification failed with error: {str(e)}")
            return (False, None)



from PySide6 import QtWidgets, QtCore
from pyqtgraph import opengl
import numpy as np
from pyvol_terminal.gl_3D_graphing.graphics_items.AbstractGLGraphicsItem import AbstractGLGraphicsItem
from pyvol_terminal.gl_3D_graphing.graphics_items.GLScatterPlotItem import GLScatterPlotItem
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GL3DViewBox



def create_objects():
    X = np.arange(-5, 5, 0.5)
    Y = np.arange(-5, 5, 0.5)
    X_mat, Y_mat = np.meshgrid(X, Y)

    m1, m2 = 2, 3 
    
    Z_mat = np.sin(X_mat) * np.cos(Y_mat)
    X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
    
    pos_scatter1 = np.column_stack((X_vec,  Y_vec, m1 * Z_vect))
    pos_scatter2 =np.column_stack((X_vec, Y_vec, m2 * Z_vect))
    
    cls = GLScatterPlotItem
    #cls = opengl.GLScatterPlotItem

    
    scatter1 = cls(pos=pos_scatter1,
                    glOptions='opaque',
                    color=(1, 0, 0, 1))

    scatter2 = cls(pos=pos_scatter2,
                    glOptions='opaque',
                    color=(1, 1, 0, 1))
    
    return scatter1, scatter2

class GLViewBox222(AbstractGLGraphicsItem, QtWidgets.QGraphicsItem):
    def __init__(self, parent=None):
        self._GLGraphicsItem__parent = None 
        self._GLGraphicsItem__children = list() 
        self._GLGraphicsItem__view = None 

        #QtWidgets.QGraphicsItem.__init__(self, None)
        super().__init__(parentItem=parent)
        
    

    def itemsChanged(self):
        print(f"\nitems have changed!!\n")


class CustomGLViewWidget(opengl.GLViewWidget):
    sigPrepareForPaint = QtCore.Signal()
    
    
    def __init__(
        self,
        parent = None,
        viewBox: GL3DViewBox | None = None,
        axisItems = None,
       # **kwargs
        ):  
        super().__init__(parent=parent)
        if viewBox is None:
            print(f"before create")
            viewBox = GL3DViewBox()
        self.vb = viewBox
        self.sigPrepareForPaint.connect(viewBox.prepareForPaint)
        super().addItem(self.vb)
        
    def addItem(self, item, ignoreBounds=False):
        if not item in self.items:
            super().addItem(item)
            if not ignoreBounds and not self.vb is None:
                self.vb.addItem(item)


class CustomGraphicsItem(AbstractGLGraphicsItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)      

    def _internal_update(self):...



class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)      
        central_widget = QtWidgets.QWidget()
        self.setCentralWidget(central_widget)
        layout = QtWidgets.QVBoxLayout(central_widget)
        self.item_object = CustomGraphicsItem()
        self.item_group = GLItemGroup(parentItem=self.item_object)
        
        #self.view_box = GLViewBox()
        #self.gl_view = CustomGLViewWidget(viewBox=self.view_box)
        self.gl_view = opengl.GLViewWidget()
        self.gl_view.setCameraPosition(distance=40)
        self.showMaximized()
        layout.addWidget(self.gl_view)
        self.gl_view.opts["distance"]=15

        return 
        
        
        
                

        
        
        
        #self.childgroup = GLItemGroup(self.view_box)   # Does not correctly normalise

        self.childitem1, self.childitem2  = create_objects()

      

        #self.childgroup.itemsChangedListeners.append(self.view_box)
        print("\nbefore add1\n")
        #self.childgroup.addItem(self.childitem1)
        print("\nbefore add2")
        #self.childgroup.addItem(self.childitem2)

        #self.gl_view.addItem(self.view_box)
        self.gl_view.addItem(self.childitem1)
        self.gl_view.addItem(self.childitem2)

        
        #self.childgroup.compute_normaliser()
        #tr = Transform3D()
       #tr.scale(0.5, 0.5, 0.5)
        #self.childgroup.setTransform(tr)
        print(self.childitem1.pos)


        for i in range(1, 11):
            x = opengl.GLTextItem(pos = (i/10, 0, 0), text=f"{round(i/10, 1)}", color="white")
            y = opengl.GLTextItem(pos = (0, i/10, 0), text=f"{round(i/10, 1)}", color="yellow")
            z = opengl.GLTextItem(pos = (i/10, 0, i/10), text=f"{round(i/10, 1)}", color="cyan")
            self.gl_view.addItem(x)
            self.gl_view.addItem(y)
            self.gl_view.addItem(z)

        

        
def main():
    
    app = QtWidgets.QApplication()
    win = Window()
    
    app.exec()


if __name__ == "__main__":
    main()



from PySide6 import QtGui
from PySide6 import QtCore, QtWidgets
from PySide6.QtOpenGLWidgets import QOpenGLWidget as OpenGLWidget
import OpenGL.GL as GL  # python wrapping of OpenGL
from OpenGL import GLU  # OpenGL Utility Library, extends OpenGL functionality
import sys  # we'll need this later to run our Qt application
import random
import numpy as np

class MainWindow(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super().__init__()

        self.resize(300, 300)
        self.setWindowTitle('Hello OpenGL App')

        width, height = 640, 480

        self.opengl = GLWidget(width, height)

        self.initGUI()

    def initGUI(self):
        self.button = QtWidgets.QPushButton('Create Triangle', self)
        self.button.clicked.connect(self.draw_triangle)

        self.sliderX = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
        self.sliderX.valueChanged.connect(self.sliderXCallBack)

        mainLayout = QtWidgets.QVBoxLayout()
        mainLayout.addWidget(self.opengl)
        mainLayout.addWidget(self.sliderX)
        mainLayout.addWidget(self.button)
        self.setLayout(mainLayout)

    def draw_triangle(self):
        gl_width, gl_height = self.opengl.size().width(), self.opengl.size().height()
        width = random.randint(50, 100)
        height = random.randint(50, 100)
        line = []
        for l in range(3):
            line.append(random.randint(0, gl_width - width))
            line.append(random.randint(0, gl_height - height))
            self.opengl.points.append(line)


    def sliderXCallBack(self):
        val = self.sliderX.value()
        pass


class GLWidget(OpenGLWidget):

    def __init__(self, width, height, parent=None):
        super().__init__(parent)
        self.setMinimumSize(width, height)
        self.init_geometry()
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_graph)
        self.timer.setInterval(0)
        self.timer.start()

    def init_geometry(self):
        self.points = [[100, 100], [200, 350], [34, 260]]


    def initializeGL(self):
        vertices = np.array([0.0, 1.0, -1.0, -1.0, 1.0, -1.0], dtype=np.float32)

        bufferId = GL.glGenBuffers(1)
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, bufferId)
        GL.glBufferData(GL.GL_ARRAY_BUFFER, vertices.nbytes, vertices, GL.GL_STATIC_DRAW)

        GL.glEnableVertexAttribArray(0)
        GL.glVertexAttribPointer(0, 2, GL.GL_FLOAT, GL.GL_FALSE, 0, None)

    # def init_ortho(self):
    #     screen_width = 800
    #     screen_height = 800
    #     ortho_left = -400
    #     ortho_right = 400
    #     ortho_top = -400
    #     ortho_bottom = 400
    #     gl.glMatrixMode(gl.GL_PROJECTION)
    #     gl.glLoadIdentity()


    def resizeGL(self, width, height):
        pass


    # def plot_point(self):
    #     gl.glBegin(gl.GL_POINTS)
    #     for p in self.points:
    #         gl.glVertex2f(p[0], p[1])
    #     gl.glEnd()

    # def plot_lines(self):
    #     localpoints = [[100, 100], [200, 350], [34, 260]]
    #     gl.glBegin(gl.GL_LINE_STRIP)
    #     gl.glVertex2d(100, 100)
    #     gl.glVertex2d(200, 350)
    #     gl.glVertex2d(34, 260)
    #     gl.glEnd()

    def paintGL(self):
        GL.glDrawArrays(GL.GL_TRIANGLES, 0, 3)


    def update_graph(self):
        self.update()


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)

    win = MainWindow()
    win.show()

    sys.exit(app.exec())


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option, Spot
    from instruments.utils import InstrumentManager
    from ...workers import SurfaceCalibration
    from ...quantities.engines import TickEngineManager
    from ...data_classes.classes import AbstractDataClass, VolatilityData, VolVector, Surface, Points, Slice
    from ...engines.interpolation_engines import Abstract3DInterpolator
    from ...engines.surface_engines import AbstractSurfaceEngine
    from ..subplot import SubPlot

from PySide6 import QtCore, QtWidgets, QtGui
from . import init_helpers
from .. import custom_widgets, subplot
import pyqtgraph as pg
from . import pyvol_GL3DViewWidget
from pyvol_terminal.misc_classes import PriceText
import time
from ...gl_3D_graphing.graphics_items import GL3DGraphicsItems
from .pyvol_GL3DGraphicsItems import PyVolGL3DSurfacePlotItem, PyVolGL3DScatterPlotItem
from pyvol_terminal.data_classes import builders as builders_data_classes
import traceback
from ..abstract_classes import ABCMainViewQSplitter
from dataclasses import dataclass, field, InitVar
import numpy as np
from pprint import pprint   


QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)

format = QtGui.QSurfaceFormat()
format.setSwapInterval(1)  
QtGui.QSurfaceFormat.setDefaultFormat(format)


class MainView(ABCMainViewQSplitter):
    def __init__(self,
                 instrument_manager: InstrumentManager,
                 tick_engine_manager: TickEngineManager=None,
                 queue_interaction: bool=False,
                 colour_styles_config: Dict[str, str]={},
                 **kwargs
                 ):
        super().__init__(QtCore.Qt.Horizontal)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        
        self.pyvol_gl_widget: pyvol_GL3DViewWidget.PyVolGL3DViewWidget=None
        self.volatility_data_container: Dict[str, VolatilityData] = kwargs.get("volatility_data_container", None)
        self.queue_interaction=queue_interaction
        
        self.interact_finish_callbacks=[]
        
        self.colour_styles_config=colour_styles_config

        all_price_types = instrument_manager.config["options"]["price_types"]
        
        self._dclass_name_subplot_title_map: Dict[str, str]={"skew" : "Skew",
                                                             "term" : "Term Structure"
                                                             }
        
        self.c=0
        self.nticks=50
        self.phase=0
        self.step=0
        
        self.display_map = DisplayMap(all_price_types)
        
        self._all_assets = ["surface", "scatter"]
        self._displayed_assets = ["surface", "scatter"]
        self._subplot_widget_container: Dict[str, SubPlot]={}
        
        self.displayed_price_types=[]
        self.displayed_price_plot_type_flags = {px_type: {"surface" : False, "scatter" : False} for px_type in all_price_types}
        self.displayed_price_plot_type_names = {px_type: {"surface" : None, "scatter" : None} for px_type in all_price_types}
        
        self.plot_state = {"surface" : True,
                           "scatter" : True,
                           "subplots" : True,
                           }

        self.scatter_from_calibrated=True

        spot_objects = [item for item in instrument_manager.spot_instrument_container.objects.values()]

        self._initViews(tick_engine_manager, spot_objects, self._dclass_name_subplot_title_map.values())
        self._initLayout()
        
    def _initViews(self,
                   tick_engine_manager: TickEngineManager,
                   spot_objects,
                   subplot_names,
                   ):
        spot_text = PriceText(spot_objects)
        
        self.pyvol_gl_widget = pyvol_GL3DViewWidget.PyVolGL3DViewWidget(spot_text, computeNormals=False)
        self.pyvol_gl_widget.processIntenseInteraction = self.processInteractionState
        
        self.addIntenseInteraction("view")
            
        axes_3D_items, axes_2D_items = init_helpers.initAxis(self.pyvol_gl_widget,
                                                             tick_engine_manager=tick_engine_manager)
        self.pyvol_gl_widget.setAxisItems(axes_3D_items)
        _all_axis="xy"
        for i, (title, axis) in enumerate(zip(subplot_names, _all_axis)):
            tick_engine = tick_engine_manager.get_engine(_all_axis[-i-1])
            axisItems={"bottom" : axes_2D_items[i],
                       "left"   : axes_2D_items[i+2]
                       }
            subplot_widget = subplot.SubPlot(title=title,
                                             axis_3D_dir=axis,
                                             other_axis_tick_engine=tick_engine,
                                             axisItems=axisItems
                                             )
            self.pyvol_gl_widget.rcMapCoordsSig.connect(subplot_widget.update3DMouseClick)
            self._subplot_widget_container[title]=subplot_widget
    
    def _initLayout(self):
        container_widget = QtWidgets.QWidget()
        surface_layout = QtWidgets.QGridLayout(container_widget)
        self.splitter_subplots = QtWidgets.QSplitter(QtCore.Qt.Vertical)  
        
        for subplot_item in self._subplot_widget_container.values():
            graphics_view = pg.GraphicsView()
            graphics_view.setCentralItem(subplot_item)
            self.splitter_subplots.addWidget(graphics_view)

        surface_layout.addWidget(self.pyvol_gl_widget, 0, 0)
        surface_layout.addWidget(self.pyvol_gl_widget.get_legend(), 0, 0, alignment=QtCore.Qt.AlignBottom | QtCore.Qt.AlignRight)
        self.addWidget(container_widget)
        self.addWidget(self.splitter_subplots)
        self.setSizes([20000, 10000])   
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        
    def set_aw(self, aw):
        self.another_window=aw
        
    def _internal_update_view(self, calibrated_surface_container: Dict[str, SurfaceCalibration]):
            print(calibrated_surface_container)
        #if not self._intense_interaction:
            self.update_dataclasses(calibrated_surface_container)
            self.update_plots()     
            self._queued_update=False
        #else:
        #    self._queued_update=True

    def update_dataclasses(self, calibrated_surface_container: Dict[str, SurfaceCalibration]):
        for px_type in self.displayed_price_types:
            surface_engine = calibrated_surface_container[px_type]
            volatility_data = self.volatility_data_container[px_type]
            for plot_type in self._displayed_assets:
                dataclass = volatility_data.get_dataclass(plot_type)
                dataclass.evaluate_from_engine(surface_engine)
                
    

    def update_plots(self):
        for px_type in self.displayed_price_types:
            volatility_data = self.volatility_data_container[px_type]
            for plot_type in self._displayed_assets:
                dataclass = volatility_data.get_dataclass(plot_type)
                internal_id = self.displayed_price_plot_type_names[px_type][plot_type]
                
                gl_plotdataitem = self.pyvol_gl_widget.plotItem(internal_id)
                gl_plotdataitem.setData(**dataclass.plot_item_kwargs())
                                
                if plot_type == "surface":
                    for sub_dataclass in volatility_data.surface_children():
                        if sub_dataclass.valid_values:
                            subplot_title = self._dclass_name_subplot_title_map[sub_dataclass.name]
                            subplot_widget = self._subplot_widget_container[subplot_title]
                            subplot_widget.pricePlotDataItem(px_type).setData(**sub_dataclass.plot_item_kwargs())
                                                
    def _add_surface_plot_data_item(self, px_type):
        if not self.displayed_price_plot_type_flags[px_type]["surface"]: 
            volatility_data = self.volatility_data_container[px_type]            
            surface_plotdataitem = PyVolGL3DSurfacePlotItem(px_type=px_type,
                                                            colormap=self.colour_styles_config["surface"][px_type],
                                                            color=self.colour_styles_config["scatter"][px_type],
                                                            **volatility_data.surface.plot_item_kwargs()
                                                            )
            self.pyvol_gl_widget.addItem(surface_plotdataitem, ignoreBounds=False)
            
            self.displayed_price_plot_type_flags[px_type]["surface"]=True
            self.displayed_price_plot_type_names[px_type]["surface"]=surface_plotdataitem.id()
            
            if self.plot_state["subplots"]:
                self._add_subplot_line(px_type, volatility_data)

    def _add_subplot_line(self, px_type: str, volatility_data: VolatilityData):
        pen = pg.mkPen(color=self.colour_styles_config["scatter"][px_type])
        for sub_dataclass in volatility_data.surface_children():
            subplot_title = self._dclass_name_subplot_title_map[sub_dataclass.name]
            if subplot_title in self._subplot_widget_container:
                subplot_widget = self._subplot_widget_container[subplot_title]
                _ = subplot_widget.plotFromDataClass(sub_dataclass,
                                                     px_type=px_type,
                                                     pen=pen
                                                     )

    def _add_scatter_plot_data_item(self, px_type):        
        if not self.displayed_price_plot_type_flags[px_type]["scatter"]:
            volatility_data = self.volatility_data_container[px_type]
            scatter_plotdataitem = PyVolGL3DScatterPlotItem(px_type=px_type,
                                                            color=self.colour_styles_config["scatter"][px_type],
                                                            **volatility_data.scatter.plot_item_kwargs()
                                                            )
            
            self.pyvol_gl_widget.addItem(scatter_plotdataitem, ignoreBounds=False)
            
            self.displayed_price_plot_type_flags[px_type]["scatter"]=True
            self.displayed_price_plot_type_names[px_type]["scatter"]=scatter_plotdataitem.id()

    def _remove_surface_plot_data_item(self, px_type):
        if px_type in self.pyvol_gl_widget.surface_plotitems:
            self.pyvol_gl_widget.removeItem(self.pyvol_gl_widget.surface_plotitems[px_type])
            self.displayed_price_plot_type_names[px_type]["surface"]=None
            self.displayed_price_plot_type_flags[px_type]["surface"]=False
            
            if self.plot_state["subplots"]:
                volatility_data = self.volatility_data_container[px_type]

                for sub_dataclass in volatility_data.surface_children():
                    subplot_title = self._dclass_name_subplot_title_map[sub_dataclass.name]
                    if subplot_title in self._subplot_widget_container:
                        subplot_widget = self._subplot_widget_container[subplot_title]
                        subplot_widget.removePricePlotDataItem(px_type)              

    def _remove_scatter_plot_data_item(self, px_type):
        if px_type in self.pyvol_gl_widget.scatter_plotitems:
            self.pyvol_gl_widget.removeItem(self.pyvol_gl_widget.scatter_plotitems[px_type])
            self.displayed_price_plot_type_names[px_type]["scatter"]=None
            self.displayed_price_plot_type_flags[px_type]["scatter"]=False
            
    def toggle_price_type(self, flag, px_type):
        if flag:
            if not px_type in self.displayed_price_types:
                if self.plot_state["surface"]:
                    self._add_surface_plot_data_item(px_type)
                if self.plot_state["scatter"]:
                    self._add_scatter_plot_data_item(px_type)
                self.displayed_price_types.append(px_type)    
        else:
            if px_type in self.displayed_price_types:
                if self.plot_state["surface"]:
                    self._remove_surface_plot_data_item(px_type)
                if self.plot_state["scatter"]:
                    self._remove_scatter_plot_data_item(px_type)
                self.displayed_price_types.remove(px_type)

    def _toggle_surface(self, check):
        if check == self.plot_state["surface"]:
            return
        if not self.plot_state["surface"]:
            for px_type in self.pyvol_gl_widget.get_displayed_price_types().copy():
                self._add_surface_plot_data_item(px_type)
        else:
            for px_type in self.pyvol_gl_widget.get_displayed_price_types().copy():
                self._remove_surface_plot_data_item(px_type)
        self.plot_state["surface"] = not self.plot_state["surface"]
        
    def _toggle_scatter(self, check):
        if check == self.plot_state["scatter"]:
            return 
        if not self.plot_state["scatter"]:
            for px_type in self.pyvol_gl_widget.get_displayed_price_types().copy():
                self._add_scatter_plot_data_item(px_type)
        else:
            for px_type in self.pyvol_gl_widget.get_displayed_price_types().copy():
                self._remove_scatter_plot_data_item(px_type)
        self.plot_state["scatter"] = not self.plot_state["scatter"]
        
    def toggle_3D_objects(self, check, plot_type):
        if plot_type == "surface":
            self._toggle_surface(check)
        elif plot_type == "scatter":
            self._toggle_scatter(check)
             
    def toggle_subplots(self, state):
        if state=="On" and not self.plot_state["subplots"]:
            self.splitter_subplots.show()
            self.plot_state["subplots"]=True

        elif state=="Off" and self.plot_state["subplots"]:
            self.splitter_subplots.hide()
            self.plot_state["subplots"]=False
            
    def interface_switch(self, *args):
        self._internal_update_view()

    def plot_queued_update(self):
        if self._queued_update:
            self.update_plots()    
            self._queued_update=False
    
    def generate_random(self):
        pass
    
    def switch_axis(self, axis_label, axis_direction):
        if isinstance(axis_direction, str):
            str_int_map = {"x" : 0,
                           "y" : 1,
                           "z" : 2,
                           }
            axis_direction=str_int_map[axis_direction]
        for axis_item in self.pyvol_gl_widget.axes[axis_direction]:
            if axis_item.style["showLabel"]:
                axis_item.setLabel(axis_label)
            
        for subplot in self._subplot_widget_container.values():
            if subplot.axisSlice() == axis_direction:
                if axis_direction == 2:
                    axis_item = subplot.getAxis("left")
                else:
                    axis_item = subplot.getAxis("bottom")
                axis_item.setLabel(axis_label)
       # self.pyvol_gl_widget.vb.
      #  self.pyvol_gl_widget.vb.enableAutoRange(enable=True)

    def switch_axis_units(self, scale, axis):
        scale = float(scale)
        if isinstance(axis, str):
            if axis == "x":
                ax=0
            elif axis == "y":
                ax=1
            elif axis == "z":
                ax=2
        for axis_item in self.pyvol_gl_widget.axes[ax]:
            axis_item.setScale(scale)
        
        for subplot in self._subplot_widget_container.values():
            if ax == 2:
                axis_item = subplot.getAxis("left")
                axis_item.setScale(scale)
                axis_item.update()
            else:
                if subplot.axisSlice() == ax:
                    axis_item = subplot.getAxis("bottom")
                    axis_item.setScale(scale)
                    axis_item.update()       
    
    def dataRange(self):
        x_values, y_values, z_values = [], [], []
        for px_type in self.displayed_price_types:
            volatility_data = self.volatility_data_container[px_type]
            for plot_type in self._displayed_assets:
                dataclass = volatility_data.get_dataclass(plot_type)
                x, y, z = dataclass.data()
                x_values = np.append(x_values, x)
                y_values = np.append(y_values, y)
                z_values = np.append(z_values, z.flatten())
        
        x_range = np.nanmin(x_values), np.nanmax(x_values)
        y_range = np.nanmin(y_values), np.nanmax(y_values)
        z_range = np.nanmin(z_values), np.nanmax(z_values)
        return x_range, y_range, z_range
    

@dataclass(slots=True)
class DisplayMap:
    all_price_types: InitVar[List[str]]
    price_types: List[str] = field(default=list)
    plot_types: List[str] = field(default=list)
    price_plot_flag: Dict[str, Dict[str, bool]] = field(default_factory=dict)
    price_plot_name: Dict[str, Dict[str, str|None]] = field(default_factory=dict)

    def __post_init__(self, all_price_types):
        self.price_plot_flag.update({px_type: {"surface" : False, "scatter" : False} for px_type in all_price_types})
        self.price_plot_name.update({px_type: {"surface" : None, "scatter" : None} for px_type in all_price_types})
    
    def add_plots_for_price_type(self, px_type):
        plots_to_add = []
        for plot_type, flag in self.price_plot_flag[px_type].items():
            if not flag:
                self.price_plot_flag[px_type][plot_type]=True        
        return plots_to_add
    
    def remove_plots_for_price_type(self, px_type):
        plots_to_remove = []
        for plot_type, flag in self.price_plot_flag[px_type].items():
            if not flag:
                self.price_plot_flag[px_type][plot_type]=True        
     #   return plots_to_add





#%%
import numpy as np


def print_output(p, vr, new_vr):
    pct_c_d_from_point_to_lim = np.abs(p.reshape(vr.shape[0], 1) - vr) / np.abs(vr[:, :1] - vr[:, 1:])
    pct_c = (100 * (new_vr / vr -1)).round(5)
    print("\n\n")
    print(f"Point to range limit:")
    print(f"x: {pct_c_d_from_point_to_lim[0]}\ny: {pct_c_d_from_point_to_lim[1]}\nz: {pct_c_d_from_point_to_lim[2]}\n")
    print(f"pct change:")
    print(f"x: {pct_c[0]}\ny: {pct_c[1]}\nz: {pct_c[2]}\n")


def zoom_2d(point, drag, viewRange, base=1.1):
    axis_mins = np.array([vr[0] for vr in viewRange])
    axis_maxs = np.array([vr[1] for vr in viewRange])

    
    scales = base ** -drag
    
    new_axis_mins = scales * axis_mins + (1 - scales) * point
    new_axis_maxs = scales * axis_maxs + (1 - scales) * point
    
    return [[new_min, new_max] for new_min, new_max in zip(new_axis_mins, new_axis_maxs)]




drags = [np.array([0.01, 0.02, 0.015]), np.array([-0.02, 0.03, 0.012])]
viewRanges = [np.array([[10,20], [-23, 45], [-55, -40]], dtype=float), np.array([[-30,44], [-30, 23], [-40,-50]], dtype=float)]

points = [np.array([19, 10, -30]), np.array([21, 22., -43])]


for point, drag, viewRange in zip(points, drags, viewRanges):
    newviewRange = zoom_2d(point, drag, viewRange)
    
    
    
#%%

print("--- Original Test Cases ---")
drags_orig = [np.array([0.01, 0.02, 0.015]), np.array([-0.02, 0.03, 0.012])]
viewRanges_orig = [np.array([[10,20], [-23, 45], [-55, -40]], dtype=float), np.array([[-30,44], [-30, 23], [-40,-50]], dtype=float)]
points_orig = [np.array([19, 10, -30]), np.array([21, 22., -43])]

for i in range(len(points_orig)):
    point = points_orig[i]
    drag = drags_orig[i]
    viewRange = viewRanges_orig[i]
    newviewRange = zoom_2d(point, drag, viewRange)
    #print_output(point, viewRange, newviewRange)

print("\n--- Edge Case Tests ---")

# Edge Case 1: Zero drag
point_ec1 = np.array([15, 0, -47.5])
drag_ec1 = np.array([0.0, 0.0, 0.0])
viewRange_ec1 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
print("\nTest Case: Zero Drag")
newviewRange_ec1 = zoom_2d(point_ec1, drag_ec1, viewRange_ec1)
#print_output(point_ec1, viewRange_ec1, newviewRange_ec1)

# Edge Case 2: Large positive drag (zoom in significantly)
point_ec2 = np.array([15, 0, -47.5])
drag_ec2 = np.array([5.0, 5.0, 5.0])
viewRange_ec2 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
print("\nTest Case: Large Positive Drag (Zoom In)")
newviewRange_ec2 = zoom_2d(point_ec2, drag_ec2, viewRange_ec2)
#print_output(point_ec2, viewRange_ec2, newviewRange_ec2)

# Edge Case 3: Large negative drag (zoom out significantly)
point_ec3 = np.array([15, 0, -47.5])
drag_ec3 = np.array([-5.0, -5.0, -5.0])
viewRange_ec3 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
print("\nTest Case: Large Negative Drag (Zoom Out)")
newviewRange_ec3 = zoom_2d(point_ec3, drag_ec3, viewRange_ec3)
#print_output(point_ec3, viewRange_ec3, newviewRange_ec3)

# Edge Case 4: Point at min of view range
point_ec4 = np.array([10, -10, -50])
drag_ec4 = np.array([0.1, 0.1, 0.1])
viewRange_ec4 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
print("\nTest Case: Point at Min of View Range")
newviewRange_ec4 = zoom_2d(point_ec4, drag_ec4, viewRange_ec4)
#print_output(point_ec4, viewRange_ec4, newviewRange_ec4)

# Edge Case 5: Point at max of view range
point_ec5 = np.array([20, 10, -45])
drag_ec5 = np.array([0.1, 0.1, 0.1])
viewRange_ec5 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
print("\nTest Case: Point at Max of View Range")
newviewRange_ec5 = zoom_2d(point_ec5, drag_ec5, viewRange_ec5)
#print_output(point_ec5, viewRange_ec5, newviewRange_ec5)

# Edge Case 6: Point outside view range (below min)
point_ec6 = np.array([5, -20, -60])
drag_ec6 = np.array([0.1, 0.1, 0.1])
viewRange_ec6 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
print("\nTest Case: Point Outside View Range (Below Min)")
newviewRange_ec6 = zoom_2d(point_ec6, drag_ec6, viewRange_ec6)
#print_output(point_ec6, viewRange_ec6, newviewRange_ec6)

# Edge Case 7: Point outside view range (above max)
point_ec7 = np.array([25, 20, -30])
drag_ec7 = np.array([0.1, 0.1, 0.1])
viewRange_ec7 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
print("\nTest Case: Point Outside View Range (Above Max)")
newviewRange_ec7 = zoom_2d(point_ec7, drag_ec7, viewRange_ec7)
#print_output(point_ec7, viewRange_ec7, newviewRange_ec7)

# Edge Case 8: View range with min and max being the same (zero width)
# This will likely cause division by zero in print_output for pct_c_d_from_point_to_lim
point_ec8 = np.array([5, 5, 5])
drag_ec8 = np.array([0.1, 0.1, 0.1])
viewRange_ec8 = np.array([[5, 5], [5, 5], [5, 5]], dtype=float)
print("\nTest Case: Zero Width View Range (Expect Division by Zero for 'Point to range limit')")
newviewRange_ec8 = zoom_2d(point_ec8, drag_ec8, viewRange_ec8)
#print_output(point_ec8, viewRange_ec8, newviewRange_ec8)

# Edge Case 9: View range with negative values and crossing zero
point_ec9 = np.array([0, 0, 0])
drag_ec9 = np.array([0.1, 0.1, 0.1])
viewRange_ec9 = np.array([[-10, 10], [-5, 5], [-20, -10]], dtype=float)
print("\nTest Case: View Range with Negative Values and Crossing Zero")
newviewRange_ec9 = zoom_2d(point_ec9, drag_ec9, viewRange_ec9)
#print_output(point_ec9, viewRange_ec9, newviewRange_ec9)

# Edge Case 10: Different base for zoom_2d
point_ec10 = np.array([15, 0, -47.5])
drag_ec10 = np.array([0.1, 0.1, 0.1])
viewRange_ec10 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
base_ec10 = 2.0
print(f"\nTest Case: Different Base ({base_ec10})")
newviewRange_ec10 = zoom_2d(point_ec10, drag_ec10, viewRange_ec10, base=base_ec10)
#print_output(point_ec10, viewRange_ec10, newviewRange_ec10)

# Edge Case 11: Base < 1 (e.g., 0.5)
point_ec11 = np.array([15, 0, -47.5])
drag_ec11 = np.array([0.1, 0.1, 0.1])
viewRange_ec11 = np.array([[10, 20], [-10, 10], [-50, -45]], dtype=float)
base_ec11 = 0.5
print(f"\nTest Case: Base < 1 ({base_ec11})")
newviewRange_ec11 = zoom_2d(point_ec11, drag_ec11, viewRange_ec11, base=base_ec11)
#print_output(point_ec11, viewRange_ec11, newviewRange_ec11)

# Edge Case 12: Point and view range with all zeros
point_ec12 = np.array([0, 0, 0])
drag_ec12 = np.array([0.1, 0.1, 0.1])
viewRange_ec12 = np.array([[0, 0], [0, 0], [0, 0]], dtype=float)
print("\nTest Case: All Zeros (Expect Division by Zero for 'Point to range limit')")
newviewRange_ec12 = zoom_2d(point_ec12, drag_ec12, viewRange_ec12)
#print_output(point_ec12, viewRange_ec12, newviewRange_ec12)





# %%


test_cases = [
    (np.array([0, 0]), np.array([0, 0]), np.array([[1, 2], [3, 4]], dtype=float)),
    (np.array([5, 5]), np.array([5, 5]), np.array([[0, 10], [0, 10]], dtype=float)),
    (np.array([5, 5]), np.array([-5, -5]), np.array([[0, 10], [0, 10]], dtype=float)),
    (np.array([1000, -1000]), np.array([1, -1]), np.array([[0.0001, 0.0002], [999.9999, 1000.0001]], dtype=float)),
    (np.array([5, 5]), np.array([0.5, -0.5]), np.array([[5, 5], [5, 5]], dtype=float)),
    (np.array([0, 0]), np.array([1e-10, -1e-10]), np.array([[-1e10, 1e10], [-1e-10, 1e-10]], dtype=float)),
    (np.array([-1, 1]), np.array([2, 2]), np.array([[0, 0], [0, 0]], dtype=float)),
    (np.array([0.5, 0.5]), np.array([1, 1]), np.array([[0, 1], [0, 1]], dtype=float)),
    (np.array([0, 0]), np.array([10, -10]), np.array([[1, 2], [3, 4]], dtype=float)),
    (np.array([1e5, -1e5]), np.array([0.1, 0.2]), np.array([[1e5-1, 1e5+1], [-1e5-1, -1e5+1]], dtype=float))
]

for i, (point, drag, viewRange) in enumerate(test_cases):
    result = zoom_2d(point, drag, viewRange)
    print(f"Test case {i+1}:")
    print("Result:", result)


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.widgets.view_box import ViewBox

from PySide6 import QtWidgets, QtCore
import sys
from pyvol_terminal.gl_3D_graphing.widgets import GL3DViewWidget
from pyqtgraph import opengl
import numpy as np
import uuid 
from pyvol_terminal.gl_3D_graphing.graphics_items import GL3DAxisItem


class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        central = QtWidgets.QWidget()
        self.setCentralWidget(central)
        self.central_layout = QtWidgets.QVBoxLayout()
        central.setLayout(self.central_layout)
        self.gl_view = GL3DViewWidget.GL3DViewWidget(padding=.2)
        X = np.arange(-5, 5, 0.5)
        Y = np.arange(-5, 5, 0.5)
        
      #  X = (X -X.min()) / (X.max() - X.min())
     #   Y = (Y -Y.min()) / (Y.max() - Y.min())
        
        X_mat, Y_mat = np.meshgrid(X, Y)
        
        R_shifted = np.sqrt((X_mat)**2 + (Y_mat)**2)
        Z_mat = 2 + np.sin(R_shifted)                
        X_vec, Y_vec, Z_vect = X_mat.flatten(), Y_mat.flatten(), Z_mat.flatten()
        
        pos_scatter = np.column_stack((X_vec, Y_vec, Z_vect))
        pos_surface = [X, Y, Z_mat]

        
   #     scatter = opengl.GLScatterPlotItem(pos=pos_scatter, color=(1, 0, 0, 1))
        surface = opengl.GLSurfacePlotItem(*pos_surface,
                                    glOptions='opaque',
                                    color=(0.5, 0.5, 1, 1))
       # self.gl_view.addItem(scatter)
        self.gl_view.addItem(surface)
        self.axis_item = GL3DAxisItem.AxisGrid(0,
                                       1,
                                       0.1,
                                       ww=self
                                       )
        self.axis_item.linkToView(self.gl_view.vb)
      #  self.axis_item.linkGLWidget(self)

     #   self.scatter=scatter
        self.surface=surface
        
        #scatter = CustomScatter(pos=pos)
        
        
        self.central_layout.addWidget(self.gl_view)     
    #    z_new = surface._z * 2
     #   surface.setData(z=z_new)
        #self.gl_view.addItem(scatter, ignoreBounds=False) 

        for i in range(1, 11):
            #x = opengl.GLTextItem(pos = (i/10, 0, 0), text=f"{round(i/10, 1)}", color="white")
            y = opengl.GLTextItem(pos = (0, i/10, 0), text=f"{round(i/10, 1)}", color="yellow")
            z = opengl.GLTextItem(pos = (i/10, 0, i/10), text=f"{round(i/10, 1)}", color="cyan")
            #self.gl_view.addItem(x, ignoreBounds=True)
            self.gl_view.addItem(y, )#ignoreBounds=True)
            self.gl_view.addItem(z, )#ignoreBounds=True)
            
        self.showMaximized()
        
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update)
        #self.timer.start(500)
        
    def update(self):
        i,j=np.unravel_index(np.argmax(self.surface._z), self.surface._z.shape)
        z_new= self.surface._z.copy()
        z_new[i,j] = z_new[i,j]+0.5
        self.surface.setData(z=z_new)
        
        pos = self.scatter.pos.copy()
        idx = np.argmax(pos[:,0])
        pos[idx,0] = pos[idx,0]+0.1
        self.scatter.setData(pos=pos)
        #self.gl_view.vb.normalize_scene()
        


def main():
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()
    

    



import sys, json, time, asyncio
from multiprocessing import Process
import numpy as np
import websockets
from PySide6.QtWidgets import QApplication, QLabel, QVBoxLayout, QWidget
import atexit
from pyqtgraph import opengl as gl
import qasync  # pip install qasync

async def server_handler(ws):
    while True:
        await ws.send(json.dumps({"price": float(np.random.rand()), "ts": time.time()}))
        await asyncio.sleep(0.5)

async def run_server():
    async with websockets.serve(server_handler, "localhost", 8766):
        await asyncio.Future()

def server_process():
    asyncio.run(run_server())

class MainWindow(QWidget):
    def __init__(self):
        super().__init__()
        self.label = QLabel("waiting...")
        layout = QVBoxLayout(self)
        layout.addWidget(self.label)

    def update_from_ws(self, data):
        self.label.setText(f"price={data['price']:.4f} ts={data['ts']:.2f}")

async def ws_listener(window):
    async with websockets.connect("ws://localhost:8766") as ws:
        while True:
            msg = await ws.recv()
            window.update_from_ws(json.loads(msg))



class GLScatterDemo(QWidget):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("GLScatter + Orbit + LocalSocket updates")
        self.resize(900, 700)

        layout = QVBoxLayout(self)
        self.status = QLabel("not connected")
        layout.addWidget(self.status)

        self.view = gl.GLViewWidget()
        self.view.opts['distance'] = 20
        layout.addWidget(self.view, 1)

        self.max_points = 1024
        self.points = np.zeros((0, 3), dtype=float)
        self.sizes = np.array([], dtype=float)
        self.colors = np.empty((0,4), dtype=float)

        self.scatter = gl.GLScatterPlotItem(pos=np.array([[0,0,0]]), size=8, color=(1,1,1,1))
        self.view.addItem(self.scatter)

        self._angle = 0.0

        g = gl.GLGridItem()
        g.scale(2,2,1)
        self.view.addItem(g)
        self.view.setCameraPosition(azimuth=0.0, elevation=20, distance=20)
        self.show()


if __name__ == "__main__":
    p = Process(target=server_process, daemon=True)
    p.start()
    atexit.register(lambda: p.terminate())

    app = QApplication(sys.argv)
    loop = qasync.QEventLoop(app)
    asyncio.set_event_loop(loop)
    scatter = GLScatterDemo()
    win = MainWindow()
    win.show()

    loop.create_task(ws_listener(win))
    with loop:
        loop.run_forever()



import warnings
import traceback

def custom_warning_handler(message, category, filename, lineno, file=None, line=None):
    """Custom warning handler that prints full traceback."""
    print(f"\n{category.__name__}: {message} ({filename}, line {lineno})")
    print("Full traceback (most recent call first):")
    traceback.print_stack()

warnings.showwarning = custom_warning_handler



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GL3DViewBox
    from pyvol_terminal.gl_3D_graphing.widgets.GL3DViewWidget import GL3DViewWidget

import pyqtgraph.opengl as gl

from pyqtgraph.opengl import GLGraphicsItem, GLViewWidget
from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
import weakref
from pyqtgraph.Qt import isQObjectAlive
from pyvol_terminal.gl_3D_graphing.meta import QABCMeta, abc
import pyqtgraph as pg
from pyqtgraph import getConfigOption, functions as fn
from .GL3DTextItem import GL3DTextItem
from .GL3DGraphicsItems import GL3DLinePlotDataItem, MixedGL3DTextItem
from .GL3DGraphicsItemMixin import GL3DGraphicsItemMixin
from .GL3DPlotDataItemMixin import GL3DFlatMeshPlotDataItemMixin
from .GL3DGraphicsItems import MixedGLLinePlotItem
import math
import copy
from pprint import pprint
from ... import utils 
from OpenGL import GLU, GL


from pyqtgraph.Qt import QtGui, QT_LIB
import importlib


class TickItem(MixedGLLinePlotItem):
    def __init__(self, tickLength, axis=None, face=None, name=None, *args, **kwargs):
        self.tickLength=tickLength
        self.axis=axis
        self.face=face
        self.name=name
        super().__init__(*args, **kwargs)
        self.setFlag(self.GLGraphicsItemFlag.ItemHasNoContents)


    
class TickLineset(MixedGLLinePlotItem):
    def __init__(self,
                 *args,
                 **kwargs
                 ):
        
        mode = "lines"    # Keep mode lines for optimized painting for grouped line items 
        kwargs["mode"]=mode
        
        self._lineItems: List[TickItem]=[]
        self._linkedLinesets: List[TickItem]=[]
        
        self._contantArray: np.ndarray=None
        
        self._lineItemsContantMap: Dict[TickItem, int] = {}
        self._linkedLinesetsConstantMap: Dict[TickLineset, Dict[TickItem, Tuple[int, int]]] = {}
        
        super().__init__(*args, **kwargs)

        
    def addConstantLineItem(self, lineItem: TickItem):
        pos = np.array(lineItem.pos)
        
        if self._contantArray is None:
            self._contantArray = pos
            self._lineItemsContantMap[lineItem] = (0, pos.shape[0])
        else:
            self._lineItemsContantMap[lineItem] = (self._contantArray.shape[0], pos.shape[0])
            self._contantArray = np.vstack((self._contantArray, pos))            

    def tickLinesets(self):
        return list(self._linkedLinesetsConstantMap.keys())
    
    def linkTickLineset(self, tickLineset: TickLineset):
        
        #if len(self.tickLinesets()) == 0:
        #    self._linkedLinesetsConstantMap = {tickLineset : {}}
        
        for lineItem in tickLineset.constantLineItems():
            pos = np.array(lineItem.pos)
            if self._contantArray is None:
                self._contantArray = pos
                self._linkedLinesetsConstantMap[tickLineset] = {lineItem : (0, pos.shape[0])}
            else:
                self._linkedLinesetsConstantMap[tickLineset] = {lineItem : (self._contantArray.shape[0], pos.shape[0])}
                self._contantArray = np.vstack((self._contantArray, pos))      
                
        self._linkedLinesets.append(tickLineset)
    
    def linkedLinesets(self) -> List[TickLineset]:
        return self._linkedLinesets
            
    def constantLineItems(self):
        return list(self._lineItemsContantMap.keys())
    
    def lineItems(self):
        return self._lineItems

    def addLineItem(self, item: MixedGLLinePlotItem):        
        self._lineItems.append(item)
    
    def childrenBoundingRect(self):...
    
    def _getWorldPositions(self, item: TickItem):
        """Get item positions in world coordinates"""
        if not hasattr(item, 'pos') or item.pos is None:
            return None
      
        tr = item.transform()
        # Apply transform to get world positions
        pos = item.pos.copy()
        if not tr.isIdentity():
            points_h = np.hstack([pos, np.ones((pos.shape[0], 1))])
            tr_matrix = np.array(tr.data()).reshape(4, 4).T
            transformed_h = points_h @ tr_matrix.T
            pos = transformed_h[:, :3]
        return pos
    
    def stackLineset(self, newPositionAll, lineItems: List[TickItem]):
        for item in lineItems:
            if item.transform().isIdentity():
                newPos = item.pos
            else:
                tr = item.transform()
                tr_matrix = np.array(tr.data()).reshape(4, 4).T  
                points_h = np.hstack([item.pos, np.ones((item.pos.shape[0], 1))])
                transformed_h = points_h @ tr_matrix.T
                newPos = transformed_h[:, :3]
            
            if newPositionAll is None:
                newPositionAll=newPos
            else:
                newPositionAll = np.vstack((newPositionAll, newPos))
        return newPositionAll
    
    def _extractNestedLines(self, combined_pos):
        T = (self.parentItem().transform() * self.transform()).inverted()[0]
        for nestedLineItem in self.linkedLinesets():
            for lineItem in nestedLineItem.lineItems():
                pos = [T.map(QtGui.QVector3D(p[0], p[1], p[2])) for p in lineItem.pos]
                
                pos = [nestedLineItem.parentItem().transform().map(QtGui.QVector3D(p.x(), p.y(), p.z())) for p in pos]
                pos = [nestedLineItem.transform().map(QtGui.QVector3D(p.x(), p.y(), p.z())) for p in pos]
                pos = np.array([[p.x(), p.y(), p.z()] for p in pos])
                
                pos = self._mapLineItem(pos, lineItem.transform())
                combined_pos = pos if combined_pos is None else np.vstack((combined_pos, pos))
                
            self.last_child2pos = pos[-1]
        return combined_pos                    
            
            
    def _mapLineItem(self, pos, tr):        
        if not tr.isIdentity():
            points_h = np.hstack([pos, np.ones((pos.shape[0], 1))])
            tr_matrix = np.array(tr.data()).reshape(4, 4).T
            transformed_h = points_h @ tr_matrix.T
            #pos = transformed_h[:, :3]
            pos = transformed_h[:, :3] / transformed_h[:, 3:4]  
        return pos
    
    def _paintHelper(self):
        newPositionAll = self.stackLineset(self._contantArray, self.lineItems())
        #for lineItem in self.lineItems():
        #    pos = self._mapLineItem(lineItem.pos.copy(), lineItem.transform())
        #    newPositionAll = pos if newPositionAll is None else np.vstack((newPositionAll, pos))
                

        newPositionAll = self._extractNestedLines(newPositionAll)
       
        if not newPositionAll is None:
            self.blockUpdates(True)
            self.setData(pos=newPositionAll, color=self.color)
            self.blockUpdates(False)
            
        super()._paintHelper()
        
        
    
    def paint(self):
        if not (GL3DGraphicsItemMixin.GLGraphicsItemFlag.ItemHasNoContents & GL3DGraphicsItemMixin.flags(self)): 
            self._paintHelper()
            self._childPaint()
        
        

class GL3DAxisTextItem(GL3DTextItem):
    
    def __init__(self,
                 tickLevel: float=None,
                 ref: int=None,
                 **kwargs
                 ):        
        
        super().__init__(**kwargs)
        self._idRef=ref
        
        self.setTickLevel(tickLevel)
        
        
    def setTickLevel(self, tickLevel):self._tickLevel=tickLevel
    def setOffset(self, offset):self._offset=offset
    
    def tickLevel(self):return self._tickLevel
    def offset(self):return self._offset
    def idRef(self):return self._idRef
    
    def setPos(selfm, kwargs):
        return kwargs

    
    @QtCore.Slot(object)
    def updateTextVisibility(self, axis_item: GL3DViewBox):
        if self.idRef() is None:
            return
        if self.idRef() > axis_item._currentTicks[self._tickLevel]-1:
            self.hide()
        else:
            self.show()

class GL3DAxisLabelItem(GL3DAxisTextItem):
    def __init__(self, *args, **kwargs):
        self._linkedTextItems: List[GL3DAxisTextItem]=[]
        super().__init__(*args, **kwargs)
    
    def linkTextValueItems(self, items: List[GL3DAxisTextItem]):
        self._linkedTextItems=items
    
    def _updatePositionFromAnchor(self, pos: QtGui.QVector3D):
        x, y = pos.x(), pos.y()
        font_metrics = QtGui.QFontMetrics(self.font)
        self.lastRect = font_metrics.tightBoundingRect(self.text)

        height = self.lastRect.height()
        width = self.lastRect.width()
        x = x - self._anchor[0] * width if self._anchor[0] == 1 else x
        y = y + self._anchor[1] * height if self._anchor[1] == 1 else y
        
        
        self.lastRect.moveTo(x, y)
        
        # X-direction offset (adjust as needed)
        
        x_offset = 5  
        max_attempts = 20
        attempts = 0
        px_xpos_collect=[]
        px_xposEnd_collect=[]
        for tick in self._linkedTextItems:
            if tick.visible():
                px_xpos_collect.append(tick.posF.x())
                px_xposEnd_collect.append(tick.posF.x() + tick.width if self.anchor()[0]==0 else tick.posF.x() - tick.width)
        n_ticks = len(px_xpos_collect)
        mid_xpos =None
        mid_xposEnd = None
        if n_ticks > 0:
            if n_ticks % 2 == 0:
                mid_xpos = 0.5 * (px_xpos_collect[int(n_ticks/2)-1] + px_xpos_collect[int(n_ticks/2)])
                mid_xposEnd = 0.5 * (px_xposEnd_collect[int(n_ticks/2)-1] + px_xposEnd_collect[int(n_ticks/2)])
            else:
                mid_xpos = px_xpos_collect[int(n_ticks/2)]
                mid_xposEnd = px_xposEnd_collect[int(n_ticks/2)]
            
            
                
            if self._anchor[0] == 0:
                if mid_xposEnd > x:
                    x += mid_xposEnd - x + 5 
            else:
                if mid_xposEnd < x:
                    x += - (x - mid_xposEnd) -5
        pos.setX(x)
        pos.setY(y)
    
                

class GL3DAxisItem(GL3DGraphicsItemMixin):
    sigAxisChanged = QtCore.Signal(QtCore.QObject)
    
    
    """
    
    axis indexes are ordered x == 0, y == 1, z == 2
       
    """
    
    
    directionAxisStr = {0 : "x", 1 : "y", 2: "z"}
    
    
    def __init__(self,
                 axis: int,
                 axisFace: int,
                 worldOrigin: Tuple[int, int, int]=(0, 0, 0),
                 viewRange: List[List[int]]=None,
                 faceRectSize: Tuple[int, int, int]=None,                 
                 text: str="",
                 labelFont: QtGui.QFont | None=QtGui.QFont("Neue Haas Grotesk", 16),
                 labelColor: str=getConfigOption('foreground'),
                 labelOffset: float=0.3, 
                 labelResizeToWorld: bool=True,
                 labelAutoUpdateAnchor: bool=True,
                 labelAnchor: Tuple[float, float]=(0.5, 0.5),
                 showLabel: bool=True,
                 valuesFont: QtGui.QFont | None=QtGui.QFont("Neue Haas Grotesk", 10),
                 valuesColor: Tuple[float,...]=getConfigOption('foreground'),
                 tickLength: float=None,
                 tickColor: Tuple[float,...]=getConfigOption('foreground'),
                 linkView: GL3DViewBox=None,
                 parentItem=None,
                 showFaceBorder: bool=True,
                 showValues: bool=True,
                 faceTickOffset: float=-0.05,
                 tickTextOffset: float=0.03,
                 tickMarginOffset: float=0,
                 graphFaceOffset: float=-0.05,
                 invertedFromWorld: bool=False,
                 valuesAnchor: Tuple[float, float]=(0., 0.),
                 textResizeToWorld: bool=True,
                 textAutoUpdateAnchor: bool=True,
                 syncToView: bool=True,
                 showGrid: bool=True,
                 showTicks: bool=True,
                 showTickMargin: bool=True,
                 **kwargs
                 ):
        super().__init__(parentItem=parentItem)
        self.flipped=False

        self._initLayout(axis, axisFace, worldOrigin, viewRange)
        self._initPerformanceParameters()
        
        self.setParentItem(parentItem)
        
        self._nTicksMajor = 6
        
        self._linkedView = None
        self._name=None
        if tickLength is None:
            tickLength = abs([size for ax, size in enumerate(faceRectSize) if ax != axis and size != 0].pop(0)) / 14
        
        self.style = {
            #        'tickTextOffset': [5, 2],  ## (horizontal, vertical) spacing between text and axis
                    'tickTextWidth': 30,  ## space reserved for tick text
                    'tickTextHeight': 18,
                    'autoExpandTextSpace': True,  ## automatically expand text space if needed
                    'autoReduceTextSpace': True,
                #   'hideOverlappingLabels': hide_overlapping_labels,
                    'valuesFont': valuesFont,
                    "valuesColor" : valuesColor,
                    "tickColor" : tickColor,
                    'stopAxisAtTick': (False, False),  ## whether axis is drawn to edge of box or to last tick
                    'textFillLimits': [  ## how much of the axis to fill up with tick text, maximally.
                        (0, 0.8),    ## never fill more than 80% of the axis
                        (2, 0.6),    ## If we already have 2 ticks with text, fill no more than 60% of the axis
                        (4, 0.4),    ## If we already have 4 ticks with text, fill no more than 40% of the axis
                        (6, 0.2),    ## If we already have 6 ticks with text, fill no more than 20% of the axis
                                    ],
                    'showValues': showValues,
                    "showFaceBorder"  : showFaceBorder,
                    "tickLength" : tickLength,
                    "maxNticksperLevel" : 10,
                    "tickWidth" : [2.5, 1.5],
                    'maxTickLevel': 1,
                    'maxTextLevel': 1,
                    'tickAlpha': None,  ## If not none, use this alpha for all ticks.
                    "invertedFromWorld" : int(invertedFromWorld),
                    "valuesAnchor" : valuesAnchor,
                    "textAutoUpdateAnchor" : textAutoUpdateAnchor,
                    "textAnchorVar" : None,
                    "tickTextOffset" : tickTextOffset,
                    "labelOffset" : labelOffset,
                    "faceTickOffset" : faceTickOffset,
                    "graphFaceOffset" : graphFaceOffset,
                    "showGrid" : showGrid,
                    "showLabel" : showLabel,
                    "showTicks" : showTicks,
                    "showTickMargin" : showTickMargin,
                    "syncToView" : syncToView,
                    
                }
        self.textWidth = 30  ## Keeps track of maximum width / height of tick text
        self.textHeight = 18
        self._currentTicks=[0 for _ in range(self.style["maxTickLevel"] + 1)]
        
        self.transformCondtions = {}
        
        self.offsetStyle = {"graphFaceOffset" : graphFaceOffset,
                            "tickMargin" : tickMarginOffset,
                            "ticks" : faceTickOffset,
                           # "tickText" : ValuesOffset,
                            "labelOffsetFace" : labelOffset,
                            }
    
        self.fixedWidth = None
        self.fixedHeight = None

        self.logMode = False

        self._tickDensity = 1.0   # used to adjust scale the number of automatically generated ticks
        self._tickLevels  = None  # used to override the automatic ticking system with explicit ticks
        self._tickSpacing = None  # used to override default tickSpacing method
       # self.scale = 1.0
        self.scale_pyqtgraph=1
        self.autoSIPrefix = False
        self.autoSIPrefixScale = 1.0
        self._siPrefixEnableRanges = None    
        

        self.labelText = ""
        self.labelUnits = ""
        self.labelUnitPrefix = ""
        self._labelItem = None
        
        self.label = None
        self.tickText = None
        
        self._GLItems = {"values" : None,
                         "ticks" : None, 
                         "tickLineset" : None,
                         "label" : None,
                         "faceBorder" : None,
                         "tickMargin" : None,
                         }
        self._GLItems: Dict[str, MixedGL3DTextItem|GL3DLinePlotDataItem]
        
        self.tickStyle = {"textResizeToWorld" : textResizeToWorld,
                          "valuesAnchor" : valuesAnchor,
                          "textAutoUpdateAnchor" : textAutoUpdateAnchor or valuesAnchor != (0, 0),
                          "showValues" : showValues,
                          'valuesFont': valuesFont,
                          "tickLength" : tickLength,
                          }
        
        labelAutoUpdateAnchor = showLabel and (labelAutoUpdateAnchor or labelAnchor != (0, 0))
        self.labelStyle = {"labelText" : text,
                           "labelColor" : labelColor or getConfigOption('foreground'),
                           "labelFont" : labelFont,
                           "labelAnchor" : labelAnchor,
                           "labelAutoUpdateAnchor" : labelAutoUpdateAnchor,
                           "labelResizeToWorld" : labelResizeToWorld,
                           "labelOffsetFace" : labelOffset,
                           "showLabel" : showLabel,
                           }
        self._cameraPositionChanged=True
        self._baseTickWorldCoords = None
        
        self._initGLItems()
    
        self.setRange(*self.viewBoxRange("axis"))
        
        if self.style["showLabel"]:
            self.setLabel(**self.labelStyle)

        self._linkedView = None
        
        if linkView is not None:
            self._linkToView_internal(linkView)
            
        self.setTickStyle(valuesColor=valuesColor,
                          valuesFont=QtGui.QFont("Neue Haas Grotesk", 10),
                          valuesAnchor=self.style["valuesAnchor"],
                          )
        
        self._setItemData()    
    
    
    def _initGLItems(self):
        self._GLItems["tickLineset"] = self._initTickItems()
        
        if self.style["showValues"]:
            self.tickText = self._initValueTextItems()
        
        if self.style["showLabel"]:
            self.label = self._initLabelTextItem()        
            self.label.linkTextValueItems(self.tickText[0])    
        
        if self.style["showFaceBorder"]:
            self._GLItems["faceBorder"] = self._initFaceRectBorderItem()     
          
        if self.style["showTicks"] and self.style["showTickMargin"]:
            self._GLItems["tickMargin"] = self._initTickMarginItem()
        
                
    def _setDataHelper(self):...
    def clipDataFromVRange(self):...
    
    def enableAutoSIPrefix(self, enable=True):
        """
        Enable (or disable) automatic SI prefix scaling on this axis.

        When enabled, this feature automatically determines the best SI prefix
        to prepend to the label units, while ensuring that axis values are scaled
        accordingly.

        For example, if the axis spans values from -0.1 to 0.1 and has units set
        to 'V' then the axis would display values -100 to 100
        and the units would appear as 'mV'

        This feature is enabled by default, and is only available when a suffix
        (unit string) is provided to display on the label.

        Parameters
        ----------
        enable : bool, optional
            Enable Auto SI prefix, by default True.
        """

        self.autoSIPrefix = enable
        self.updateAutoSIPrefix()
    
    
    def updateAutoSIPrefix(self):
        scale = 1.0
        prefix = ''
        if self.label.visible():
            _range = 10**np.array(self.range) if self.logMode else self.range
            scaling_value = max(abs(_range[0]), abs(_range[1])) * self.scale_pyqtgraph
            if any(low <= scaling_value <= high for low, high in self.getSIPrefixEnableRanges()):
                (scale, prefix) = fn.siScale(scaling_value)

        self.autoSIPrefixScale = scale
        self.labelUnitPrefix = prefix
        
    def _updateLabel(self):
        if self.style["showLabel"]:
            self.label.setData(text=self.labelString())
        
    def getSIPrefixEnableRanges(self):
        """
        Get the ranges in which automatic SI prefix scaling is enabled.

        Returns
        -------
        tuple of tuple of float, float
            A tuple of ranges where SI prefix scaling is enabled. Each range is a tuple
            containing two floats representing the start and end of the range. If no
            custom ranges are set, then the default ranges are returned. The default
            ranges are ``((0., 1.), (1e9, inf))`` if units are empty, and 
            ``((0., inf))`` otherwise.
        """
        if self._siPrefixEnableRanges is not None:
            return self._siPrefixEnableRanges
        elif self.labelUnits == '':
            return (0., 1.), (1e9, float('inf'))
        else:
            return ((0., float('inf')),)
        
    def viewBoxRange(self, ax_or_ax_type):
        if isinstance(ax_or_ax_type, int):
            return self.geometry["viewRange"][ax_or_ax_type]
        elif isinstance(ax_or_ax_type, str):
            return self.geometry["viewRange"][self.geometry[ax_or_ax_type]]
    
    def _transformToView(self, viewRange):
        xmin, xmax = viewRange[0]
        ymin, ymax = viewRange[1]
        zmin, zmax = viewRange[2]
        self.blockUpdates(True)
        self.scale(xmax - xmin, ymax - ymin, zmax - zmin, False)
        self.blockUpdates(False)
        self.translate(xmin, ymin, zmin, False)

    
    def _linkToView_internal(self, view: GL3DViewBox):
        self.unlinkFromView()
        self._linkedView = weakref.ref(view)
        self._transformToView(view.worldRange())

        if self.geometry["axis"] == 0:
            view.sigXRangeChanged.connect(self.linkedViewChanged)
        elif self.geometry["axis"] == 1:
            view.sigYRangeChanged.connect(self.linkedViewChanged)
        else:
            view.sigZRangeChanged.connect(self.linkedViewChanged)
        view.sigResized.connect(self.linkedViewChanged)
        
        if isinstance(view.parentWidget(), GLViewWidget):
            plotitem_widget = view.parentWidget()
                        
            if self.style["syncToView"]:
                plotitem_widget.sigViewAngleChanged.connect(self.cameraPositionChanged)
                self.autoUpdatePosition(plotitem_widget)
        
            if self.tickStyle["textAutoUpdateAnchor"] or self.labelStyle["labelAutoUpdateAnchor"]:
                    plotitem_widget.sigViewAngleChanged.connect(self.cameraPositionChanged)
                    self.autoUpdateAnchor(plotitem_widget)
                
    def _initTransform(self, origin):
        tr = self.transform()
        for ax in range(3):
            
            translation = np.zeros(3)
            translation[ax] = -0.5 

            tr.translate(* (-1 *translation))

            scales = [1, 1, 1]
            scales[ax] *= (1 - 2 * origin[ax])
            
            tr.scale(*scales)
            tr.translate(*translation)
        self.setTransform(tr)


    def _initLayout(self, axis, axisFace, worldOrigin, viewRange):
        if viewRange is None:
            viewRange = np.repeat(np.array([[0. ,1.]]), 3, axis=0)
        else:
            viewRange = np.array(viewRange, dtype=float)
            
        self.geometry={"axis" : axis,
                       "axisFace" : axisFace,
                       "axisOrtho" : ({0, 1, 2} - {axis, axisFace}).pop(),
                       "localOrigin" : [0., 0., 0.],
                       "worldOrigin" : worldOrigin,
                       "viewRange" : np.repeat(np.array([[0. ,1.]]), 3, axis=0),  # set dummy viewRange before calling  _setViewRange
                       }
        
        self._initTransform(worldOrigin)
        self.geometry["viewRange"] = viewRange
    
    def _initPerformanceParameters(self):
        self.__geometryPerformance = {}
        
        
        z_1 = self.geometry["localOrigin"].copy()
        z_1[self.geometry["axisFace"]] = np.diff(self.viewBoxRange("axisFace")).item()
        
        self.__geometryPerformance["localRangeQVector"] = pg.Vector(*self.geometry["localOrigin"]), pg.Vector(*z_1)

        
    
    @classmethod
    def fromViewBoxRange(cls,
                         viewRange: list[list[float]] | Dict[str, list[float]],   # limits of the viewBox in world coordinates [[x_min, x_max], [y_min, y_max], [z_min, z_max]] or {"x" : [x_min, x_max], "y" : [y_min, y_max], "z" : [z_min, z_max]}
                         axis: str | int,
                         worldOrigin: Tuple[int, int, int]|Dict[str, int]=None, #    extremum values must be either 0 or 1, and orients the axis origin in world coordinates. {"x" : int, "y" : int ,..} or (int, int ,int)
                         faceDirection: str | int | None = None,         # faceDirection is the direction that shares the rectangular-face with the axis-direction along the viewRange. e.g. for the unit-cube with direction==x faceDirection==y for the xy face
                         **kwargs
                         ) -> 'GL3DViewBox':
        
        worldOrigin = list(worldOrigin)
        
        direction_size = [d - worldOrigin[axis] for d in viewRange[axis] if d != worldOrigin[axis]][0]
        faceDirection_size = [d - worldOrigin[faceDirection] for d in viewRange[faceDirection] if d != worldOrigin[faceDirection]][0]
        
        faceRectSize = [0, 0, 0]
        faceRectSize[axis] = direction_size
        faceRectSize[faceDirection] = faceDirection_size

        axisItem  = cls(axis,
                        axisFace=faceDirection,
                        worldOrigin=worldOrigin,
                        viewRange=viewRange,
                        faceRectSize=faceRectSize,
                        faceDirection=faceDirection,
                        **kwargs)    
        return axisItem
    
    def _initFaceRectBorderItem(self) -> opengl.GLLinePlotItem:
        
        p1 = [0.]*3
        p1[:] = self.geometry["localOrigin"]
        p2 = p1.copy()
        
        p2[self.geometry["axis"]] = 1. - self.geometry["localOrigin"][self.geometry["axis"]]
        
        p3 = p1.copy()
        p3[self.geometry["axis"]] = 1. - self.geometry["localOrigin"][self.geometry["axis"]]
        p3[self.geometry["axisFace"]] = 1. - self.geometry["localOrigin"][self.geometry["axisFace"]]
        
        p4 = p1.copy()
        p4[self.geometry["axisFace"]] = 1.- self.geometry["localOrigin"][self.geometry["axisFace"]]
        
        segments = [p1, p2, p2, p3, p3, p4, p4, p1]
        
        borderline = MixedGLLinePlotItem(pos=segments,
                                        color= self.style["tickColor"],
                                        glOptions="translucent",
                                        mode="lines",
                                        )
        borderline.name = "borderline"
        self._GLItems["tickLineset"].addConstantLineItem(borderline)
        
        borderline.setParentItem(self._GLItems["tickLineset"])
        
        return borderline
    
    
    def _initTickMarginItem(self):
        p1 = [0.]*3
        p1[:] = self.geometry["localOrigin"]
        
        p2 = p1.copy()
        p2[self.geometry["axis"]] = 1. - self.geometry["localOrigin"][self.geometry["axis"]]
        
        marginline = MixedGLLinePlotItem(pos=[p1, p2],
                                        color= self.style["tickColor"],
                                        glOptions="translucent",
                                        mode="lines",
                                        )
        marginline.setTransform(self.tickItems.transform())
        marginline.name="margin"
        self._GLItems["tickLineset"].addConstantLineItem(marginline)
        
        marginline.setParentItem(self._GLItems["tickLineset"])
        return marginline
    

    def _validate_kwargs(self, direction, orthoCoords, perpendicular_axis, viewBoxRange):
        c_viewBoxRange = viewBoxRange.copy() if isinstance(viewBoxRange, list) else viewBoxRange.tolist().copy()
        _ = c_viewBoxRange.pop(direction)
        orthoCoords = list(orthoCoords)
        if not all(orthoCoords[idx] in ax for idx, ax in enumerate(c_viewBoxRange)):
            raise ValueError("Orthogonal Coordinates do not match the viewBox range")

    def _childPaint(self):...
    
    def _setItemData(self):
        tickLevels, tickLevelsWorld = self.tickValues(*self.range, 1)
        self.updateLineItems(tickLevelsWorld)
        
        if self.style["showValues"]:
            self.updateTextItems(tickLevels, tickLevelsWorld)

    def _paintHelper(self):
        v = self.view()
        if not v is None:
            if self.style["syncToView"] and self._cameraPositionChanged:
                self.autoUpdatePosition(v, False)
            if self.style["showValues"] and self._cameraPositionChanged and (self.tickStyle["textAutoUpdateAnchor"] or self.labelStyle["labelAutoUpdateAnchor"]):
                self.autoUpdateAnchor(v, False)
            self._cameraPositionChanged=False
        self._setItemData()        
        
        self._updateLabel()
        
        self.sigAxisChanged.emit(self)

    def _initValueTextItems(self) -> List[List[GL3DAxisTextItem]]:
        gl_objects_level = []
        idx=0
        if self.style["showTicks"]:
            parent = self.tickItems
            parent_offset = parent.tickLength
        else:
            parent = self
            parent_offset = 0.
        
        for jdx in range(self.style["maxTextLevel"] + 1):
            gl_objects = []
            for idx in range(self.style["maxNticksperLevel"]):                
                gl_object = GL3DAxisTextItem(tickLevel=jdx,
                                             ref=idx,
                                             anchor=self.tickStyle["valuesAnchor"],
                                             font=self.tickStyle["valuesFont"], 
                                             resizeToWorld=True,
                                             resizeTolerance=0.1,
                                             )
                offset = [0., 0., 0.,]
                offset[self.geometry["axisFace"]] = parent_offset + self.style["tickTextOffset"]
                gl_object.translate(*offset)
                gl_object.setParentItem(parent)
                gl_object.hide()                
                self.sigAxisChanged.connect(gl_object.updateTextVisibility)
                gl_objects.append(gl_object)
                idx+=1
            gl_objects_level.append(gl_objects)
        return gl_objects_level
    
    
    
    def tickValues(self, minVal:float, maxVal:float, size: float):
        minVal, maxVal = sorted((minVal, maxVal))

        minVal *= self.scale_pyqtgraph
        maxVal *= self.scale_pyqtgraph

        ticks = []
        ticksWorld = []
        
        
        tickLevels = self.tickSpacing(minVal, maxVal, size)
        allValues = np.array([])
        
        for i in range(len(tickLevels)):
            spacing, offset = tickLevels[i]

            ## determine starting tick
            start = (math.ceil((minVal-offset) / spacing) * spacing) + offset

            ## determine number of ticks
            num = int((maxVal-start) / spacing) + 1
            values = (np.arange(num) * spacing + start) / self.scale_pyqtgraph
            ## remove any ticks that were present in higher levels
            ## we assume here that if the difference between a tick value and a previously seen tick value
            ## is less than spacing/100, then they are 'equal' and we can ignore the new tick.
            close = np.any(
                np.isclose(
                    allValues,
                    values[:, np.newaxis],
                    rtol=0,
                    atol=spacing/self.scale_pyqtgraph*0.01
                ),
                axis=-1
            )
            values = values[~close]                
            allValues = np.concatenate([allValues, values])
            ticks.append((spacing/self.scale_pyqtgraph, values.tolist()))
            ticksWorld.append((self.mapTextToWorld(spacing/self.scale_pyqtgraph), self.mapTextToWorld(values).tolist()))
            self._currentTicks[i] = len(values.tolist())
        return ticks, ticksWorld
    
    def tickStrings(self, values: list[float], scale: float, spacing: float):
        """
        Return the strings that should be displayed at each tick value.

        This method is used to generate tick strings, and is called automatically.

        Parameters
        ----------
        values : list of float
            List of tick values.
        scale : float
            The scaling factor for tick values.
        spacing : float
            The spacing between ticks.

        Returns
        -------
        list of str
            List of strings to display at each tick value.
        """
        
            
        disp_spacing = spacing * scale
        if disp_spacing > 0:
            places = max(0, math.ceil(-math.log10(disp_spacing)))
        else:
            places = 0

        strings = []
        for v in values:
            # Apply scale only once
            vs = v * scale
            if abs(vs) < .001 or abs(vs) >= 10000:
                vstr = "%g" % vs
            else:
                vstr = ("%%0.%df" % places) % vs
            strings.append(vstr)
        return strings


    def _invalidateTickCache(self):
        """Reset all cached calculations that depend on scale"""
        self._tickLevels = None
        self._autoRangeNeedsUpdate = True
        self._matrixNeedsUpdate = True
        self._initDistance = None
        self._initPointSize = None
        self._lastSetDistance = None

    def updateLineItems(self, tickLevelsWorld):
        def _append_update(base, lineLength, target_list):
            pos = base.copy()
            pos[self.geometry["axisFace"]] = lineLength
            target_list.append(base)
            target_list.append(pos)

        tickPositions = []
        gridPositions = []
        
        for idx, ticksWorld in enumerate(tickLevelsWorld, start=1):
            for tick in ticksWorld[1]:
                basePosition = [0., 0., 0., ]
                basePosition[self.geometry["axis"]] = tick
                if self.style["showTicks"]:
                    _append_update(basePosition, self.style["tickLength"] / idx, tickPositions)

                if self.style["showGrid"]:
                    _append_update(basePosition, np.diff(self.viewBoxRange("axisFace")).item(), gridPositions)

        if self.style["showTicks"]:
            self.tickItems.blockUpdates(True)
            self.tickItems.setData(pos=np.array(tickPositions))
            self.tickItems.blockUpdates(False)
            
        if self.style["showGrid"]:
            self._GLItems["Grids"].blockUpdates(True)
            self._GLItems["Grids"].setData(pos=np.array(gridPositions))
            self._GLItems["Grids"].blockUpdates(False)
            

    def updateTextItems(self, tickLevels, tickLevelsWorld):
        for jdx, (ticks, ticksWorld) in enumerate(zip(tickLevels, tickLevelsWorld)):
            spacing, values = ticks
            strings = self.tickStrings(values, self.scale_pyqtgraph, spacing)
            for idx, (value, string, tickWorld) in enumerate(zip(values, strings, ticksWorld[1])):
                
                if jdx > self.style["maxTextLevel"]:
                    break
                text_item=self.tickText[jdx][idx]
                basePosition=self.geometry["localOrigin"].copy()
                basePosition[self.geometry["axis"]] = tickWorld

                text_item.blockUpdates(True)
                text_item.setData(pos=basePosition, text=string)
                text_item.blockUpdates(False)
                
        
                
    def setRange(self, mn: float, mx: float):
        if not math.isfinite(mn) or not math.isfinite(mx):
            raise ValueError(f"Not setting range to [{mn}, {mx}]")
        
        self.range = [mn, mx]
        if self.autoSIPrefix:
            # XXX: Will already update once!
            self.updateAutoSIPrefix()
    #    self._setItemData()
        
        #self.sigAxisChanged.emit(self)        
        
        
    def setTextFont(self, font=None):
        if font is None:
            getConfigOption("foreground")

    def setScale(self, scale=1.0):
        """
        Set the value scaling for this axis.

        Setting this value causes the axis to draw ticks and tick labels as if the view
        coordinate system were scaled.

        Parameters
        ----------
        scale : float, optional
            Value to scale the drawing of ticks and tick labels as if the view
            coordinate system was scaled, by default 1.0.
        """
        
        if scale != self.scale_pyqtgraph:
            self.scale_pyqtgraph = scale
            self._updateLabel()
        self._invalidateTickCache()  
        self.update()
            
    def setLabel(self,
                labelText: str | None=None,
                labelColor: str | None=None,
                labelFont: str | None=None,
                labelAnchor: str=None,
                labelResizeToWorld: bool=None,
                labelAutoUpdateAnchor: bool=None,
                showLabel: bool=None,
                unitPrefix: str | None=None,
                **kwargs
                ):  
        if not labelText is None:
            self.labelText = labelText
        
        visible = bool(self.labelText)
        if visible:
            self.label.setData(text=self.labelText)
        else:
            self.label.setData(text="")
        if not showLabel is None:
            if showLabel and not visible:
                raise
            
        self.showLabel(visible)
        
        if "labelOffsetFace" in kwargs:
            del kwargs["labelOffsetFace"]

        if not labelColor is None:
            self.labelStyle["labelColor"] = labelColor
        if not labelFont is None:
            self.labelStyle["labelFont"] = labelFont
        if not labelAnchor is None:
            self.labelStyle["labelAnchor"] = labelAnchor
        if not labelResizeToWorld is None:
            self.labelStyle["labelResizeToWorld"] = labelResizeToWorld and visible
            self.labelText
        
        if not labelAutoUpdateAnchor is None:
            self.labelStyle["labelAutoUpdateAnchor"] = (labelAutoUpdateAnchor or labelAnchor != (0, 0)) and visible
        
        kwargs = {"resizeToWorld" : self.labelStyle["labelResizeToWorld"],
                  "text" : self.labelText,
                  "color" : self.labelStyle["labelColor"],
                  "font" : self.labelStyle["labelFont"],
             #     "anchor" : self.labelStyle["labelAnchor"],
                  }
        
        self.label.setData(**kwargs)
    
    def showLabel(self, show: bool=True):
        if show:
            self.label.show()
        else:
            self.label.hide()
    
    def setTickStyle(self,
                     tickColor: str | None=None,
                     tickAnchor: Tuple[float, float]=(0., 0.),
                     valuesFont: str | None=None,
                     textResizeToWorld: bool=True,
                     textAutoUpdateAnchor: bool=None,
                     showValues: bool=None,
                     *args,
                     **kwargs
                     ):
        
        if not tickColor is None:
            self.tickStyle["tickColor"] = tickColor            
        if not valuesFont is None:
            self.tickStyle["valuesFont"] = valuesFont            
        if not tickAnchor is None:
            self.tickStyle["tickAnchor"] = tickAnchor
        if not textResizeToWorld is None:
            self.tickStyle["textResizeToWorld"] = textResizeToWorld
        if not tickAnchor is None:
            self.tickStyle["tickAnchor"] = tickAnchor
        if not showValues is None:
            self.tickStyle["showValues"] = showValues
        
        if not textAutoUpdateAnchor is None:
            self.tickStyle["textAutoUpdateAnchor"] = (textAutoUpdateAnchor or tickAnchor != (0, 0)) and self.tickStyle["showValues"]

        if self.tickStyle["showValues"]:
        
            if "values" in self._GLItems and valuesFont:
                for tick_level in self.tickText:
                    for item in tick_level:
                        item.setData(font=self.tickStyle["valuesFont"])
    
    def _initLabelTextItem(self) -> GL3DAxisLabelItem:
        pos = np.zeros(3)
        pos[self.geometry["axis"]] = 0.5 * self.viewBoxRange("axis").sum()
        pos[self.geometry["axisFace"]] = self.style["tickLength"] + self.style["tickTextOffset"] + self.style["labelOffset"]
        labelItem = GL3DAxisLabelItem(pos=pos,
                                      anchor=self.labelStyle["labelAnchor"],
                                      font=self.labelStyle["labelFont"], 
                                      resizeToWorld=True,
                                      resizeTolerance=0.1,
                                      )
        labelItem.setParentItem(self.tickItems)
        return labelItem

    def _createTickitems(self):
            ticks = TickItem(self.style["tickLength"],
                             axis=self.geometry["axis"],
                             face=self.geometry["axisFace"],
                             glOptions="translucent",
                             antialias=True,
                             mode="lines",
                             )
            
            self.tickItems = ticks
            
            translation = np.array([0., 0., 0.])
            translation[self.geometry["axisFace"]] = self.geometry["localOrigin"][self.geometry["axisFace"]]
            
            tr = ticks.transform()
            
            tr.translate(*(-1 *translation))
            scale = [1, 1, 1]
            scale[self.geometry["axisFace"]] *= -1 
            tr.scale(*scale)

            tr.translate(*translation)
            translate_back = [0., 0., 0.,]
            translate_back[self.geometry["axisFace"]] = - 1*np.array(self.style["faceTickOffset"])
            
            tr.translate(*translate_back)
            return ticks
        
    def _initTickItems(self) -> TickLineset:
        tickItem = TickLineset(glOptions="translucent",
                                antialias=True,
                                color=self.style["tickColor"],
                                width=self.style["tickWidth"][0],
                                mode="lines",
                                )
        
        offset = [0., 0., 0.,]
        offset[self.geometry["axisOrtho"]] = self.style["graphFaceOffset"]
        tickItem.translate(*offset)
        if self.style["showTicks"]:
            ticks = self._createTickitems()
            ticks.name="ticks"
            tickItem.addLineItem(ticks)
            ticks.setParentItem(tickItem)
        
        if self.style["showGrid"]:
            gridLength = np.diff(self.viewBoxRange("axisFace")).item()
            grids = TickItem(gridLength,
                             axis=self.geometry["axis"],
                             face=self.geometry["axisFace"],
                             glOptions="translucent",
                             antialias=True,
                             mode="lines",
                             )
            grids.name="grids"
            tickItem.addLineItem(grids)
            grids.setParentItem(tickItem)
            self._GLItems["Grids"] = grids
        else:
            self._GLItems["Grids"] = None
            
        tickItem.setParentItem(self)
        return tickItem
    
    def checkWidths(self, textNum: str):...
    
    @QtCore.Slot(object)
    @QtCore.Slot(object, object)
    def linkedViewChanged(self, view: GL3DViewBox, newRange=None):
        if newRange is None:
            newRange = view.viewRange()[self.viewBoxRange("axis")]
        self.setRange(*newRange)
        

    def tickSpacing(self, minVal: float, maxVal: float, size: float):
        _tickDensity=1
        dif = abs(maxVal - minVal)
        if dif == 0:
            return []

        ref_size = 300. # axes longer than this display more than the minimum number of major ticks
        minNumberOfIntervals = max(
            2.25,       # 2.0 ensures two tick marks. Fudged increase to 2.25 allows room for tick labels.
            2.25 * _tickDensity * math.sqrt(size/ref_size) # sub-linear growth of tick spacing with size
        )

        majorMaxSpacing = dif / minNumberOfIntervals

        mantissa, exp2 = math.frexp(majorMaxSpacing)
        p10unit = 10. ** (
            math.floor(
                (exp2-1)
                / 3.32192809488736
            ) - 1
        )
        
        if 100. * p10unit <= majorMaxSpacing:
            majorScaleFactor = 10
            p10unit *= 10.
        else:
            for majorScaleFactor in (50, 20, 10):
                if majorScaleFactor * p10unit <= majorMaxSpacing:
                    break
        
        majorInterval = majorScaleFactor * p10unit
        levels = [
            (majorInterval, 0),
        ]

        if self.style['maxTickLevel'] >= 1:
            # Modified this section to choose a smaller interval for minor ticks
            if majorScaleFactor == 10:
                minorScaleFactor = 5  # Half of major interval for minor ticks
            elif majorScaleFactor == 20:
                minorScaleFactor = 10  # Half of major interval
            elif majorScaleFactor == 50:
                minorScaleFactor = 25  # Half of major interval
            else:
                minorScaleFactor = majorScaleFactor / 2  # Default to half
            
            minorInterval = minorScaleFactor * p10unit
            levels.append((minorInterval, 0))

        if self.style['maxTickLevel'] >= 2:
            # Keep the extra level logic but ensure it's smaller than minor interval
            if majorScaleFactor == 10:
                extraScaleFactor = 2  # Smaller than minor interval
            elif majorScaleFactor == 20:
                extraScaleFactor = 5  # Smaller than minor interval
            elif majorScaleFactor == 50:
                extraScaleFactor = 10  # Smaller than minor interval
            else:
                extraScaleFactor = minorScaleFactor / 2  # Default to half
            
            extraInterval = extraScaleFactor * p10unit
            if extraInterval < minorInterval:
                levels.append((extraInterval, 0))
        return levels
    
    def _axisSizePixels(self):
        view = self.view()
        
        ax_local_s, ax_local_f = pg.Vector(0, 0, 0), pg.Vector(0, 0, 1)
        
        ax_view_s, ax_view_f = self.mapToView(ax_local_s), self.mapToView(ax_local_f)
        
        
        px_x, px_y, px_z = GL3DAxisItem.map_3D_coords_to_2D(view, ax_view_s.x(), ax_view_s.y(), ax_view_s.z())



    def mapTextToWorld(self, value):
        rmin, rmax = self.range
        view = self.linkedView()
        
        if view is None:
            world_bounds = 0, 1
        else:
            world_bounds = view.state["worldRange"][self.geometry["axis"]]
                                    
        dif = rmax - rmin
        
        if rmax == rmin:
            scale = 1
            offset = 0
            value
        else:
            scale = (world_bounds[1] - world_bounds[0]) / dif
            offset = world_bounds[0] - rmin * scale

        mapped = scale * value + offset
        return mapped * (1 / (world_bounds[1] - world_bounds[0]))
    
    def labelString(self):
        if self.labelUnits == '':
            if not self.autoSIPrefix or self.autoSIPrefixScale == 1.0:
                units = ''
            else:
                units = f'(x{1.0 / self.autoSIPrefixScale:g})'
        else:
            units = f'({self.labelUnitPrefix}{self.labelUnits})'
        
        return f'{self.labelText} {units}' if bool("") else self.labelText
    
    def flipAboutOrtho(self):
        # Flip face to the other side of the graph
        tr = self.transform()
        translate = [0, 0, 0]
        translate[self.geometry["axisOrtho"]] = 1 
        tr.translate(*translate)
        scale = [1, 1, 1]
        scale[self.geometry["axisOrtho"]] *= -1
        tr.scale(*scale)

    def flipAboutFace(self):
        # Flip ticks and values to the other side of the face 
        midPoint = [0, 0, 0]
        midPoint[self.geometry["axisFace"]] = 0.5 
        midPoint_reverse = midPoint.copy()
        midPoint_reverse[self.geometry["axisFace"]] = -0.5
        
        tr = self.transform()
        tr.translate(*midPoint)

        scales = [1, 1, 1]
        scales[self.geometry["axisFace"]] *= -1
        
        tr.scale(*scales)
        tr.translate(*(midPoint_reverse))
    
    @QtCore.Slot(object)
    def cameraPositionChanged(self, plotitem_widget):
        self._cameraPositionChanged=True
    
    @QtCore.Slot(object)
    def autoUpdatePosition(self, plotitem_widget: GL3DViewWidget, block=True):
        if block:
            return
        cameraPosition = plotitem_widget.cameraPosition()
        
        mapped = self.mapFromParent(cameraPosition)        
        
        axisFacePosition = getattr(mapped, self.directionAxisStr[self.geometry["axisFace"]])()
        axisFaceRange = self.viewBoxRange("axisFace")
        
        if axisFacePosition - (axisFaceRange[1] - axisFaceRange[0]) > 0: 
            self.flipAboutFace()

        axisOrthoPosition = getattr(mapped, self.directionAxisStr[self.geometry["axisOrtho"]])()
        axisOrthoRange = self.viewBoxRange("axisOrtho")
        
        if axisOrthoPosition + (axisOrthoRange[1] - axisOrthoRange[0]) < 0: 
            self.flipAboutOrtho()
            
            
    @classmethod
    def clockwise_angle(cls, p1, p2, k2):
        v1 = p2 - p1
        v2 = k2 - p1

        dot = np.dot(v1, v2)
        det = v1[0]*v2[1] - v1[1]*v2[0]  

        angle_rad = np.arctan2(det, dot) 
        clockwise_angle = (-angle_rad) % (2 * np.pi)
        return np.degrees(clockwise_angle)
            
    @QtCore.Slot(object)
    def autoUpdateAnchor(self, plotitem_widget: GL3DViewWidget, block=True):
        if block:
            return
        pos_world_z0 = self.mapToParent(self.__geometryPerformance["localRangeQVector"][0])
        pos_world_z1 = self.mapToParent(self.__geometryPerformance["localRangeQVector"][1])
        
        camera_array0 = np.array([pos_world_z0.x(), pos_world_z0.y(), pos_world_z0.z()])
        camera_array1 = np.array([pos_world_z1.x(), pos_world_z1.y(), pos_world_z1.z()])
        
        camera_world = plotitem_widget.cameraPosition()
        camera_world = np.array([camera_world.x(), camera_world.y()])
        
        if self.clockwise_angle(camera_array0[:2], camera_array1[:2], camera_world[:2]) < 180:
            newAnchor = (1, self.style["valuesAnchor"][1])
        else:
            newAnchor = (0, self.style["valuesAnchor"][1])
        
        if self.labelStyle["labelAutoUpdateAnchor"]:
            self.label.setAnchor(newAnchor)
            
        if self.tickStyle["textAutoUpdateAnchor"]:
            if newAnchor != self.style["valuesAnchor"]:
                self.style["valuesAnchor"]=newAnchor
                for text_item_container in self.tickText:
                    for text_item in text_item_container:
                        text_item.setAnchor(newAnchor)

    def linkToView(self, view):
        self._linkToView_internal(view)

    def unlinkFromView(self):
        oldView = self.linkedView()
        self._linkedView = None
        if oldView is not None:
            oldView.sigResized.disconnect(self.linkedViewChanged)
            if self.geometry["axis"] == 0:
                oldView.sigXRangeChanged.disconnect(self.linkedViewChanged)
            elif self.geometry["axis"] == 1:
                oldView.sigYRangeChanged.disconnect(self.linkedViewChanged)
            else:
                oldView.sigZRangeChanged.disconnect(self.linkedViewChanged)

    def linkedView(self) -> GL3DViewBox:
        return None if self._linkedView is None else self._linkedView()

    def setTickFont(self, font: QtGui.QFont | None):
        self.style['tickFont'] = font
        self.picture = None
        #self.prepareGeometryChange()
        self.update()
    
        
    def linkOptimizeAxes(self, childAxis: GL3DViewBox):
        childLineset = childAxis._GLItems["tickLineset"]
        self._GLItems["tickLineset"].linkTickLineset(childLineset)
        childLineset.setFlag(childLineset.GLGraphicsItemFlag.ItemHasNoContents, True)
        childAxis.hide()
        self.update()
        
    @classmethod
    def map_3D_coords_to_2D(cls, view, world_x, world_y, world_z):
        device_pixel_ratio = view.window().screen().devicePixelRatio()
        widget_height = view.height()
        widget_width = view.width()
        
        viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
        _, _, viewport_width, viewport_height = viewport
        
        modelview = np.array(view.viewMatrix().data()).reshape(4, 4)
        projection = np.array(view.projectionMatrix(viewport,
                                                    (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio * widget_height)
                                                    ).data()
                              ).reshape(4, 4)
        
        px_x, px_y, px_z = GLU.gluProject(world_x, world_y, world_z, modelview, projection, viewport)
        px_x = viewport_width - px_x
        px_y = viewport_height - px_y
        
        return px_x, px_y, px_z




from PySide6.QtCore import QObject, Signal, QTimer
from PySide6.QtWidgets import QApplication, QMainWindow, QLabel
import multiprocessing as mp
import sys
import time
 
def worker_process(queue):
    for i in range(5):
        time.sleep(1)
        queue.put(f"Message {i} from process")

class SignalEmitter(QObject):
    message_received = Signal(str)

class MainWindow(QMainWindow):
    def __init__(self, queue):
        super().__init__()
        self.label = QLabel("Waiting for messages...", self)
        self.setCentralWidget(self.label)
        self.emitter = SignalEmitter()
        self.emitter.message_received.connect(self.label.setText)
        self.queue = queue
        self.timer = QTimer()
        self.timer.timeout.connect(self.poll_queue)
        self.timer.start(100)

    def poll_queue(self):
        while not self.queue.empty():
            self.emitter.message_received.emit(self.queue.get())

def main():
    mp.set_start_method("spawn")
    queue = mp.Queue()
    process = mp.Process(target=worker_process, args=(queue,))
    process.start()
    app = QApplication(sys.argv)
    win = MainWindow(queue)
    win.show()
    app.exec()
    process.join()

if __name__ == "__main__":
    main()



import sys
import numpy as np
from PySide6 import QtWidgets, QtCore, QtGui
import pyqtgraph.opengl as gl
from OpenGL import GL


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        
        self.widget1 = QtWidgets.QWidget()  
        self.setCentralWidget(self.widget1)
        self.widget1.setLayout(QtWidgets.QVBoxLayout())
        self.widget2 = QtWidgets.QWidget()
        self.lw2 = QtWidgets.QHBoxLayout(self.widget2)
        
        self.gl_widget = gl.GLViewWidget()
        self.lw2.addWidget(self.gl_widget)
        print(self.gl_widget.layout())
        
        self.widget1.layout().addWidget(self.widget2)
   #     self.widget1.layout().addWidget(self.gl_widget)
        self.show()

        print(self.gl_widget.isValid())

def main():
    """Main function to run the application."""
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()

    sys.exit(app.exec())

if __name__ == "__main__":
    main()



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.utils import InstrumentManager
    from pandas import DataFrame
    from engines.surface_engines import AbstractSurfaceEngine
    from data_classes.classes import BaseDomain, Domain, OptionChain, VolatilityData, VolVector
    from examples.synthetic.websocket_streamer import Streamer
    
from PySide6 import QtCore, QtWebSockets
import asyncio
import queue
from pyvol_terminal.instruments.instruments import ABCInstrument
import time
import numpy as np
from abc import ABC, abstractmethod
from .gl_3D_graphing.meta import QABCMeta
import json

def get_updater_dict():
    return {""}

class WebsocketWorker(QtCore.QObject):
    processedSignal = QtCore.Signal(dict)
    
    def __init__(self,
                 data_processing_config,
                 ws_worker_config, 
                 all_instruments: Dict[str, ABCInstrument],
                 active_instruments: set[str]=set(),
                 **kwargs
                 ):
        super().__init__()
        self._should_stop=False
        self._intense_interaction=False
        self._intense_interaction_just_ended=False
        self.task=None
        self.buffer_responses={}
        self._all_instruments=all_instruments
        self.active_instruments = active_instruments
        self.data_processing_config = data_processing_config
        self.websocket = ws_worker_config.pop("websocket")
        self.ws_worker_config=ws_worker_config
        self.generator_call = getattr(self.websocket, ws_worker_config["start_ws_func_name"])
        self.ws_instrument_key = data_processing_config["websocket_json_format"]["instrument_key"]
        self.ws_bid_key = data_processing_config["websocket_json_format"]["bid_key"]
        self.ws_ask_key = data_processing_config["websocket_json_format"]["ask_key"]
        self.ws_timestamp_key = data_processing_config["websocket_json_format"]["timestamp_key"]
        
        self.setUpdateMethod(data_processing_config)
        
        self._thread = QtCore.QThread()
        self.moveToThread(self._thread)
        self._thread.started.connect(self.start_worker)
    
    def setUpdateMethod(self, data_processing_config):
        if "buffer" in data_processing_config:
            if data_processing_config["buffer"]:
                if "timer_process_data" in data_processing_config:
                    self.timer_process_data=data_processing_config["timer_process_data"]
                else:
                    self.timer_process_data=2
                self._updater_method = self.buffer_updater
            else:
                self.timer_process_data=None
                self._updater_method=self.single_update
        else:
            self.timer_process_data=None
            self._updater_method=self.single_update
    
    def start_worker(self):
        match self.ws_worker_config["parallel_type"]:
            case "async":
                self.loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self.loop)
                self.task = self.loop.create_task(self.run_async())
                try:
                    self.loop.run_until_complete(self.task)
                except asyncio.CancelledError:
                    pass
                finally:
                    self.loop.close()
            case "threading":
                self.run_threading()
                
    async def run_async(self,):
            async for message in self.generator_call():
                if self._should_stop:
                    break
                self.update_price(message)

    def stop_worker(self):
        self._should_stop = True
                    
        if self._thread.isRunning():
            self._thread.quit()
            self._thread.wait()

        match self.ws_worker_config["parallel_type"]:
            case "async":
                if self.task and not self.task.done():
                    self.task.cancel()
                
    def update_price(self, websocket_response: Dict[Any, Any]):
        ticker = websocket_response[self.ws_instrument_key]    
        if ticker in self.active_instruments:
            timestamp = websocket_response[self.ws_timestamp_key]
            
            self._all_instruments[ticker].update(self.normalize_timestamp(timestamp),
                                                 websocket_response[self.ws_bid_key],
                                                 websocket_response[self.ws_ask_key]
                                                 )
                        

    def add_active_instruments(self, instruments: set[str]):
        self.active_instruments.update(instruments)
        
    def reset_active_instruments(self, new_active_instruments: set[str]=set()):
        self.active_instruments=new_active_instruments
            
    def check_enough_time(self):
        if time.time() - self.last_process_update > self.timer_process_data:            
            return True
        else:
            return False
    
    def buffer_updater(self, websocket_response):
        self.bulk_response(websocket_response)
        if self.check_enough_time():
            self.update_price_with_buffer()
        
    def bulk_response(self, websocket_responses):
        for websocket_response in websocket_responses:
            ticker = websocket_response[self.ws_instrument_key]
            self.buffer_responses[ticker] = websocket_response

    def update_price_with_buffer(self):
        for ticker, websocket_response in self.buffer_responses.copy().items():
            self.update_price(websocket_response)
            del self.buffer_responses[ticker]
        self.buffer_responses.clear()
        self.last_process_update=time.time()
        
    def start(self):
        self._thread.start()
        self._thread.setPriority(QtCore.QThread.LowestPriority)
    
    def single_update(self, websocket_response):
        self.update_price(websocket_response)
    
    @staticmethod
    def normalize_timestamp(ts: float) -> float:
        if ts > 1e14:     
            return ts * 1e-9
        elif ts > 1e11:     
            return ts * 1e-3
        else:              
            return ts
        
    def setIntenseInteraction(self, state: bool):
        if self._intense_interaction and not state:
            self._intense_interaction_just_ended = True
        self._intense_interaction = state



class ABCCalibrationWorker(QtCore.QObject, metaclass=QABCMeta):
    calibratedSignal = QtCore.Signal(dict)
    def __init__(self,
                 *,
                 active_instruments: set[str],
                 update_timer: int,
                 calibration_engines: Dict[str, AbstractSurfaceEngine], 
                 **kwargs
                 ):
        super().__init__(**kwargs)
        self.active_instruments=active_instruments
        self._update_timer=update_timer
        self.calibration_engines=calibration_engines
        
        self._running=False
        self._intense_interaction=False
        
        self._thread = QtCore.QThread()
        self._thread.started.connect(self.start_worker)
        
    def work(self):
        while self._running:
            while self._intense_interaction:
                QtCore.QThread.msleep(200)
            self._internal_work()
            QtCore.QThread.msleep(self._update_timer)    
            
    def setIntenseInteraction(self, flag: bool):
        self._intense_interaction=flag
    
    def start_worker(self):
        self.moveToThread(self._thread)
        self._running=True
        self._thread.started.connect(self.work)
        self._thread.start()
        
    def stop_worker(self):
        self._running=False
        if self._thread.isRunning():
            self._thread.quit() 
            self._thread.wait() 
    
    def isRunning(self): return all((self._running, self._thread.isRunning()))
    
    @abstractmethod
    def _internal_work(self):...
    

class SurfaceCalibration(ABCCalibrationWorker):
    def __init__(self,
                 *,
                 option_chain_container: Dict[str, OptionChain],
                 vol_vect_container: Dict[str, VolVector],
                 **kwargs,
                 ):
        super().__init__(**kwargs)
        self.option_chain_container=option_chain_container
        self.vol_vect_container=vol_vect_container
    
    def _internal_work(self):
        valid_calibrations=[]
        for px_type, engine in self.calibration_engines.items():
            option_chain = self.option_chain_container[px_type]
            vol_vector = self.vol_vect_container[px_type]
            vol_vector.update(**option_chain.getDataKwargs())
            engine.calibrate(*vol_vector.calibration_args())
            valid_calibrations.append(engine.valid_calibration())

        if any(valid_calibrations):
            self.calibratedSignal.emit(self.calibration_engines)
            
    